# This file was autogenerated by genconsts.py
#
# LINSTOR - management of distributed storage/DRBD9 resources
# Copyright (C) 2017 - 2022  LINBIT HA-Solutions GmbH
# All Rights Reserved.
# Author: Robert Altnoeder, Roland Kammerer, Gabor Hernadi, Rene Peinthor
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
import sys
from enum import Enum

if sys.version_info > (3,):
    long = int


# ## Bits 62 - 63 (most significant 2) are reserved for the message type masks (error, warning, info)
# Bits 25 - 26 are reserved for the operation type masks (create, modify, delete)
# Bits 18 - 22 are reserved for the object type masks (node, resource, resource definition, ...)
# Bits 0  - 14 are reserved for codes ###
MASK_BITS_TYPE = long(0xC000000000000000)
MASK_ERROR = long(0xC000000000000000)
MASK_WARN = long(0x8000000000000000)
MASK_INFO = long(0x4000000000000000)
MASK_SUCCESS = long(0x0000000000000000)


# ## Operation type masks ###
MASK_BITS_OP = long(0x0000000003000000)
MASK_CRT = long(0x0000000001000000)
MASK_MOD = long(0x0000000002000000)
MASK_DEL = long(0x0000000003000000)


# ## Type masks (Node, ResDfn, Res, VolDfn, Vol, NetInterface, ...) ###
MASK_BITS_OBJ = long(0x00000000007C0000)
MASK_SCHEDULE = long(0x0000000000540000)
MASK_EXT_FILES = long(0x0000000000500000)
MASK_PHYSICAL_DEVICE = long(0x00000000004C0000)
MASK_VLM_GRP = long(0x0000000000480000)
MASK_RSC_GRP = long(0x0000000000440000)
MASK_KVS = long(0x0000000000400000)
MASK_NODE = long(0x00000000003C0000)
MASK_RSC_DFN = long(0x0000000000380000)
MASK_RSC = long(0x0000000000340000)
MASK_VLM_DFN = long(0x0000000000300000)
MASK_VLM = long(0x00000000002C0000)
MASK_NODE_CONN = long(0x0000000000280000)
MASK_RSC_CONN = long(0x0000000000240000)
MASK_VLM_CONN = long(0x0000000000200000)
MASK_NET_IF = long(0x00000000001C0000)
MASK_STOR_POOL_DFN = long(0x0000000000180000)
MASK_STOR_POOL = long(0x0000000000140000)
MASK_CTRL_CONF = long(0x0000000000100000)
MASK_SNAPSHOT = long(0x00000000000C0000)
MASK_BACKUP = long(0x0000000000080000)
MASK_REMOTE = long(0x0000000000040000)


# ## Codes ###
MASK_BITS_CODE = long(0x0000000000007FFF)


# ## Codes 1-9: success ###
CREATED = long(1 | MASK_SUCCESS)
DELETED = long(2 | MASK_SUCCESS)
MODIFIED = long(3 | MASK_SUCCESS)
PASSPHRASE_ACCEPTED = long(4 | MASK_SUCCESS)


# ## Codes 100 - 999: failures ###


# ## Codes 100 - 199: sql failures ###
FAIL_SQL = long(100 | MASK_ERROR)
FAIL_SQL_ROLLBACK = long(101 | MASK_ERROR)


# ## Codes 200-299: invalid * failures ###
FAIL_INVLD_NODE_NAME = long(200 | MASK_ERROR)
FAIL_INVLD_NODE_TYPE = long(201 | MASK_ERROR)
FAIL_INVLD_RSC_NAME = long(202 | MASK_ERROR)
FAIL_INVLD_RSC_PORT = long(203 | MASK_ERROR)
FAIL_INVLD_NODE_ID = long(204 | MASK_ERROR)
FAIL_INVLD_VLM_NR = long(205 | MASK_ERROR)
FAIL_INVLD_VLM_SIZE = long(206 | MASK_ERROR)
FAIL_INVLD_MINOR_NR = long(207 | MASK_ERROR)
FAIL_INVLD_STOR_POOL_NAME = long(208 | MASK_ERROR)
FAIL_INVLD_NET_NAME = long(209 | MASK_ERROR)
FAIL_INVLD_NET_ADDR = long(210 | MASK_ERROR)
FAIL_INVLD_NET_PORT = long(211 | MASK_ERROR)
FAIL_INVLD_NET_TYPE = long(212 | MASK_ERROR)
FAIL_INVLD_PROP = long(213 | MASK_ERROR)
FAIL_INVLD_TRANSPORT_TYPE = long(214 | MASK_ERROR)
FAIL_INVLD_TCP_PORT = long(215 | MASK_ERROR)
FAIL_INVLD_CRYPT_PASSPHRASE = long(216 | MASK_ERROR)
FAIL_INVLD_ENCRYPT_TYPE = long(217 | MASK_ERROR)
FAIL_INVLD_SNAPSHOT_NAME = long(218 | MASK_ERROR)
FAIL_INVLD_PLACE_COUNT = long(219 | MASK_ERROR)
FAIL_INVLD_FREE_SPACE_MGR_NAME = long(220 | MASK_ERROR)
FAIL_INVLD_STOR_DRIVER = long(221 | MASK_ERROR)
FAIL_INVLD_DRBD_PROXY_COMPRESSION_TYPE = long(222 | MASK_ERROR)
FAIL_INVLD_KVS_NAME = long(223 | MASK_ERROR)
FAIL_INVLD_LAYER_KIND = long(224 | MASK_ERROR)
FAIL_INVLD_LAYER_STACK = long(225 | MASK_ERROR)
FAIL_INVLD_EXT_NAME = long(226 | MASK_ERROR)
FAIL_INVLD_PROVIDER = long(227 | MASK_ERROR)
FAIL_INVLD_VLM_SIZES = long(228 | MASK_ERROR)
FAIL_INVLD_VLM_COUNT = long(229 | MASK_ERROR)
FAIL_INVLD_CONF = long(230 | MASK_ERROR)
FAIL_INVLD_SNAPSHOT_SHIPPING_SOURCE = long(231 | MASK_ERROR)
FAIL_INVLD_SNAPSHOT_SHIPPING_TARGET = long(232 | MASK_ERROR)
FAIL_NODE_HAS_USED_RSC = long(233 | MASK_ERROR)
FAIL_INVLD_REQUEST = long(234 | MASK_ERROR)
FAIL_INVLD_EXT_FILE_NAME = long(235 | MASK_ERROR)
FAIL_INVLD_EXT_FILE = long(236 | MASK_ERROR)
FAIL_INVLD_RSC_GRP_NAME = long(237 | MASK_ERROR)
FAIL_INVLD_BACKUP_CONFIG = long(238 | MASK_ERROR)
FAIL_INVLD_REMOTE_NAME = long(239 | MASK_ERROR)
FAIL_INVLD_TIME_PARAM = long(240 | MASK_ERROR)
FAIL_INVLD_SCHEDULE_NAME = long(241 | MASK_ERROR)


# ## Codes 300-399: dependency not found failures ###
FAIL_NOT_FOUND_NODE = long(300 | MASK_ERROR)
FAIL_NOT_FOUND_RSC_DFN = long(301 | MASK_ERROR)
FAIL_NOT_FOUND_RSC = long(302 | MASK_ERROR)
FAIL_NOT_FOUND_VLM_DFN = long(303 | MASK_ERROR)
FAIL_NOT_FOUND_VLM = long(304 | MASK_ERROR)
FAIL_NOT_FOUND_NET_IF = long(305 | MASK_ERROR)
FAIL_NOT_FOUND_NODE_CONN = long(306 | MASK_ERROR)
FAIL_NOT_FOUND_RSC_CONN = long(307 | MASK_ERROR)
FAIL_NOT_FOUND_VLM_CONN = long(308 | MASK_ERROR)
FAIL_NOT_FOUND_STOR_POOL_DFN = long(309 | MASK_ERROR)
FAIL_NOT_FOUND_STOR_POOL = long(310 | MASK_ERROR)
FAIL_NOT_FOUND_DFLT_STOR_POOL = long(311 | MASK_ERROR)
FAIL_NOT_FOUND_CRYPT_KEY = long(312 | MASK_ERROR)
FAIL_NOT_FOUND_SNAPSHOT_DFN = long(313 | MASK_ERROR)
FAIL_NOT_FOUND_SNAPSHOT_VLM_DFN = long(314 | MASK_ERROR)
FAIL_NOT_FOUND_SNAPSHOT = long(315 | MASK_ERROR)
FAIL_NOT_FOUND_KVS = long(316 | MASK_ERROR)
FAIL_NOT_FOUND_RSC_GRP = long(317 | MASK_ERROR)
FAIL_NOT_FOUND_VLM_GRP = long(318 | MASK_ERROR)
FAIL_NOT_FOUND_EXOS_ENCLOSURE = long(319 | MASK_ERROR)
FAIL_NOT_FOUND_EXT_FILE = long(320 | MASK_ERROR)
FAIL_NOT_FOUND_REMOTE = long(321 | MASK_ERROR)
FAIL_NOT_FOUND_BACKUP = long(322 | MASK_ERROR)
FAIL_NOT_FOUND_SCHEDULE = long(323 | MASK_ERROR)


# ## Codes 400-499: access denied failures ###
FAIL_ACC_DENIED_NODE = long(400 | MASK_ERROR)
FAIL_ACC_DENIED_RSC_DFN = long(401 | MASK_ERROR)
FAIL_ACC_DENIED_RSC = long(402 | MASK_ERROR)
FAIL_ACC_DENIED_VLM_DFN = long(403 | MASK_ERROR)
FAIL_ACC_DENIED_VLM = long(404 | MASK_ERROR)
FAIL_ACC_DENIED_STOR_POOL_DFN = long(405 | MASK_ERROR)
FAIL_ACC_DENIED_STOR_POOL = long(406 | MASK_ERROR)
FAIL_ACC_DENIED_NODE_CONN = long(407 | MASK_ERROR)
FAIL_ACC_DENIED_RSC_CONN = long(408 | MASK_ERROR)
FAIL_ACC_DENIED_VLM_CONN = long(409 | MASK_ERROR)
FAIL_ACC_DENIED_STLT_CONN = long(410 | MASK_ERROR)
FAIL_ACC_DENIED_CTRL_CFG = long(411 | MASK_ERROR)
FAIL_ACC_DENIED_COMMAND = long(412 | MASK_ERROR)
FAIL_ACC_DENIED_WATCH = long(413 | MASK_ERROR)
FAIL_ACC_DENIED_SNAPSHOT_DFN = long(414 | MASK_ERROR)
FAIL_ACC_DENIED_SNAPSHOT = long(415 | MASK_ERROR)
FAIL_ACC_DENIED_SNAPSHOT_VLM_DFN = long(416 | MASK_ERROR)
FAIL_ACC_DENIED_FREE_SPACE_MGR = long(417 | MASK_ERROR)
FAIL_ACC_DENIED_KVS = long(418 | MASK_ERROR)
FAIL_ACC_DENIED_RSC_GRP = long(419 | MASK_ERROR)
FAIL_ACC_DENIED_VLM_GRP = long(420 | MASK_ERROR)
FAIL_ACC_DENIED_SNAP_DFN = long(421 | MASK_ERROR)
FAIL_ACC_DENIED_EXT_FILE = long(422 | MASK_ERROR)
FAIL_ACC_DENIED_REMOTE = long(423 | MASK_ERROR)
FAIL_ACC_DENIED_SCHEDULE = long(424 | MASK_ERROR)


# ## Codes 500-599: data already exists failures ###
FAIL_EXISTS_NODE = long(500 | MASK_ERROR)
FAIL_EXISTS_RSC_DFN = long(501 | MASK_ERROR)
FAIL_EXISTS_RSC = long(502 | MASK_ERROR)
FAIL_EXISTS_VLM_DFN = long(503 | MASK_ERROR)
FAIL_EXISTS_VLM = long(504 | MASK_ERROR)
FAIL_EXISTS_NET_IF = long(505 | MASK_ERROR)
FAIL_EXISTS_NODE_CONN = long(506 | MASK_ERROR)
FAIL_EXISTS_RSC_CONN = long(507 | MASK_ERROR)
FAIL_EXISTS_VLM_CONN = long(508 | MASK_ERROR)
FAIL_EXISTS_STOR_POOL_DFN = long(509 | MASK_ERROR)
FAIL_EXISTS_STOR_POOL = long(510 | MASK_ERROR)
FAIL_EXISTS_STLT_CONN = long(511 | MASK_ERROR)
FAIL_EXISTS_CRYPT_PASSPHRASE = long(512 | MASK_ERROR)
FAIL_EXISTS_WATCH = long(513 | MASK_ERROR)
FAIL_EXISTS_SNAPSHOT_DFN = long(514 | MASK_ERROR)
FAIL_EXISTS_SNAPSHOT = long(516 | MASK_ERROR)
FAIL_EXISTS_EXT_NAME = long(517 | MASK_ERROR)
FAIL_EXISTS_NVME_TARGET_PER_RSC_DFN = long(518 | MASK_ERROR)
FAIL_EXISTS_NVME_INITIATOR_PER_RSC_DFN = long(519 | MASK_ERROR)
FAIL_LOST_STOR_POOL = long(521 | MASK_ERROR)
FAIL_EXISTS_RSC_GRP = long(522 | MASK_ERROR)
FAIL_EXISTS_VLM_GRP = long(523 | MASK_ERROR)
FAIL_EXISTS_OPENFLEX_TARGET_PER_RSC_DFN = long(524 | MASK_ERROR)
FAIL_EXISTS_SNAPSHOT_SHIPPING = long(525 | MASK_ERROR)
FAIL_EXISTS_EXOS_ENCLOSURE = long(526 | MASK_ERROR)
FAIL_EXISTS_REMOTE = long(527 | MASK_ERROR)
FAIL_EXISTS_SCHEDULE = long(528 | MASK_ERROR)


# ## Codes 600-699: data missing failures ###
FAIL_MISSING_PROPS = long(600 | MASK_ERROR)
FAIL_MISSING_PROPS_NETCOM_TYPE = long(601 | MASK_ERROR)
FAIL_MISSING_PROPS_NETCOM_PORT = long(602 | MASK_ERROR)
FAIL_MISSING_NETCOM = long(603 | MASK_ERROR)
FAIL_MISSING_PROPS_NETIF_NAME = long(604 | MASK_ERROR)
FAIL_MISSING_STLT_CONN = long(605 | MASK_ERROR)
FAIL_MISSING_EXT_NAME = long(606 | MASK_ERROR)
FAIL_MISSING_NVME_TARGET = long(608 | MASK_ERROR)
FAIL_NO_STLT_CONN_DEFINED = long(609 | MASK_ERROR)
FAIL_MISSING_OPENFLEX_TARGET = long(610 | MASK_ERROR)
FAIL_MISSING_EBS_TARGET = long(611 | MASK_ERROR)


# ## Codes 700-799: uuid mismatch failures ###
FAIL_UUID_NODE = long(700 | MASK_ERROR)
FAIL_UUID_RSC_DFN = long(701 | MASK_ERROR)
FAIL_UUID_RSC = long(702 | MASK_ERROR)
FAIL_UUID_VLM_DFN = long(703 | MASK_ERROR)
FAIL_UUID_VLM = long(704 | MASK_ERROR)
FAIL_UUID_NET_IF = long(705 | MASK_ERROR)
FAIL_UUID_NODE_CONN = long(706 | MASK_ERROR)
FAIL_UUID_RSC_CONN = long(707 | MASK_ERROR)
FAIL_UUID_VLM_CONN = long(708 | MASK_ERROR)
FAIL_UUID_STOR_POOL_DFN = long(709 | MASK_ERROR)
FAIL_UUID_STOR_POOL = long(710 | MASK_ERROR)
FAIL_UUID_KVS = long(711 | MASK_ERROR)


# ## Codes 800-899: number pools exhausted ###
FAIL_POOL_EXHAUSTED_VLM_NR = long(800 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_MINOR_NR = long(801 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_TCP_PORT = long(802 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_NODE_ID = long(803 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_RSC_LAYER_ID = long(804 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_SPECIAL_SATELLTE_TCP_PORT = long(805 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_SNAPSHOT_SHIPPING_TCP_PORT = long(806 | MASK_ERROR)
FAIL_POOL_EXHAUSTED_BACKUP_SHIPPING_TCP_PORT = long(807 | MASK_ERROR)


# ## Other failures ###
FAIL_EBS_COOLDOWN = long(973 | MASK_ERROR)
FAIL_NOT_ALL_UPTODATE = long(974 | MASK_ERROR)
FAIL_EVACUATING = long(975 | MASK_ERROR)
FAIL_BACKUP_UNKNOWN_CLUSTER = long(976 | MASK_ERROR)
FAIL_DEPENDEND_BACKUP = long(977 | MASK_ERROR)
FAIL_BACKUP_INCOMPATIBLE_VERSION = long(978 | MASK_ERROR)
FAIL_SNAPSHOT_NOT_UPTODATE = long(979 | MASK_ERROR)
FAIL_NOT_ENOUGH_FREE_SPACE = long(980 | MASK_ERROR)
FAIL_ONLY_ONE_ACT_RSC_PER_SHARED_STOR_POOL_ALLOWED = long(981 | MASK_ERROR)
FAIL_CRYPT_INIT = long(982 | MASK_ERROR)
FAIL_SNAPSHOT_SHIPPING_NOT_SUPPORTED = long(983 | MASK_ERROR)
FAIL_SNAPSHOT_SHIPPING_IN_PROGRESS = long(984 | MASK_ERROR)
FAIL_UNDECIDABLE_AUTOPLACMENT = long(985 | MASK_ERROR)
FAIL_PRE_SELECT_SCRIPT_DID_NOT_TERMINATE = long(986 | MASK_ERROR)
FAIL_LINSTOR_MANAGED_SATELLITE_DID_NOT_START_PROPERLY = long(987 | MASK_ERROR)
FAIL_STLT_DOES_NOT_SUPPORT_LAYER = long(988 | MASK_ERROR)
FAIL_STLT_DOES_NOT_SUPPORT_PROVIDER = long(989 | MASK_ERROR)
FAIL_STOR_POOL_CONFIGURATION_ERROR = long(990 | MASK_ERROR)
FAIL_INSUFFICIENT_REPLICA_COUNT = long(991 | MASK_ERROR)
FAIL_RSC_BUSY = long(992 | MASK_ERROR)
FAIL_INSUFFICIENT_PEER_SLOTS = long(993 | MASK_ERROR)
FAIL_SNAPSHOTS_NOT_SUPPORTED = long(994 | MASK_ERROR)
FAIL_NOT_CONNECTED = long(995 | MASK_ERROR)
FAIL_NOT_ENOUGH_NODES = long(996 | MASK_ERROR)
FAIL_IN_USE = long(997 | MASK_ERROR)
FAIL_UNKNOWN_ERROR = long(998 | MASK_ERROR)
FAIL_IMPL_ERROR = long(999 | MASK_ERROR)


# ## Codes 1000-9999: warnings ###
WARN_INVLD_OPT_PROP_NETCOM_ENABLED = long(1001 | MASK_WARN)
WARN_NOT_CONNECTED = long(1002 | MASK_WARN)
WARN_STLT_NOT_UPDATED = long(1003 | MASK_WARN)
WARN_NO_STLT_CONN_DEFINED = long(1004 | MASK_WARN)
WARN_DEL_UNSET_PROP = long(1005 | MASK_WARN)
WARN_RSC_ALREADY_DEPLOYED = long(1006 | MASK_WARN)
WARN_RSC_ALREADY_HAS_DISK = long(1007 | MASK_WARN)
WARN_RSC_ALREADY_DISKLESS = long(1008 | MASK_WARN)
WARN_ALL_DISKLESS = long(1009 | MASK_WARN)
WARN_STORAGE_ERROR = long(1010 | MASK_WARN)
WARN_NOT_FOUND_CRYPT_KEY = long(1011 | MASK_WARN)
WARN_STORAGE_KIND_ADDED = long(1012 | MASK_WARN)
WARN_NOT_ENOUGH_NODES_FOR_TIE_BREAKER = long(1013 | MASK_WARN)
WARN_MIXED_PMEM_AND_NON_PMEM = long(1014 | MASK_WARN)
WARN_UNEFFECTIVE_PROP = long(1015 | MASK_WARN)
WARN_INVLD_SNAPSHOT_SHIPPING_PREFIX = long(1016 | MASK_WARN)
WARN_NODE_EVICTED = long(1017 | MASK_WARN)
WARN_RSC_DEACTIVATED = long(1018 | MASK_WARN)
WARN_BACKUP_DL_ONLY = long(1019 | MASK_WARN)
WARN_NOT_EVACUATING = long(1020 | MASK_WARN)
WARN_INVLD_CONF = long(1021 | MASK_WARN)
WARN_NOT_FOUND = long(3000 | MASK_WARN)


# ## Codes 10000-19999: info ###
INFO_NO_RSC_SPAWNED = long(10000 | MASK_INFO)
INFO_NODE_NAME_MISMATCH = long(10001 | MASK_INFO)
INFO_PROP_SET = long(10002 | MASK_INFO)
INFO_TIE_BREAKER_CREATED = long(10003 | MASK_INFO)
INFO_TIE_BREAKER_DELETING = long(10004 | MASK_INFO)
INFO_TIE_BREAKER_TAKEOVER = long(10006 | MASK_INFO)
INFO_PROP_REMOVED = long(10005 | MASK_INFO)
INFO_AUTO_DRBD_PROXY_CREATED = long(10007 | MASK_INFO)
INFO_NOOP = long(10007 | MASK_INFO)
INFO_RSC_ALREADY_EXISTS = long(10008 | MASK_INFO)


# ## Special codes ###
UNKNOWN_API_CALL = long(0x0FFFFFFFFFFFFFFF | MASK_ERROR)
API_CALL_AUTH_REQ = long(0x0FFFFFFFFFFFFFFE | MASK_ERROR)
API_CALL_PARSE_ERROR = long(0x0FFFFFFFFFFFFFFD | MASK_ERROR)


# ## SignIn codes ###
SUCCESS_SIGN_IN = long(10000 | MASK_SUCCESS)
FAIL_SIGN_IN = long(10000 | MASK_ERROR)
FAIL_SIGN_IN_MISSING_CREDENTIALS = long(10001 | MASK_ERROR)


# ## Special answer message content types ###
# Textual MsgApiCallResponse responses
API_REPLY = str('Reply')
# Indicates that the immediate answers to the API call are complete
API_END_OF_IMMEDIATE_ANSWERS = str('EndOfImmediateAnswers')


# ## Create object APIs ###
API_CRT_NODE = str('CrtNode')
API_CRT_RSC = str('CrtRsc')
API_CRT_RSC_DFN = str('CrtRscDfn')
API_CRT_NET_IF = str('CrtNetIf')
API_CRT_VLM_DFN = str('CrtVlmDfn')
API_CRT_SNAPSHOT = str('CrtSnapshot')
API_CRT_BACKUP = str('CrtBackup')
API_CRT_STOR_POOL_DFN = str('CrtStorPoolDfn')
API_CRT_STOR_POOL = str('CrtStorPool')
API_CRT_NODE_CONN = str('CrtNodeConn')
API_CRT_RSC_CONN = str('CrtRscConn')
API_CRT_VLM_CONN = str('CrtVlmConn')
API_AUTO_PLACE_RSC = str('AutoPlaceRsc')
API_CLONE_RSCDFN = str('CloneRscDfn')
API_CLONE_RSCDFN_STATUS = str('CloneRscDfnStatus')
API_CRT_CRYPT_PASS = str('CrtCryptPass')
API_CRT_OF_TARGET_NODE = str('CrtOfTargetNode')
API_RESTORE_VLM_DFN = str('RestoreVlmDfn')
API_RESTORE_SNAPSHOT = str('RestoreSnapshot')
API_CRT_RSC_GRP = str('CrtRscGrp')
API_CRT_VLM_GRP = str('CrtVlmGrp')
API_SPAWN_RSC_DFN = str('SpawnRscDfn')
API_ADJUST_RSC_GRP = str('AdjustRscGrp')
API_CREATE_DEVICE_POOL = str('CreateDevicePool')
API_MAKE_RSC_AVAIL = str('MakeRscAvail')
API_CRT_EXOS_ENCLOSURE = str('CrtExosEnclosure')
API_SET_REMOTE = str('SetRemote')
API_RESTORE_BACKUP = str('RstBackup')
API_ABORT_BACKUP = str('AbortBackup')
API_SHIP_BACKUP = str('ShipBackup')
API_CRT_SCHEDULE = str('CrtSchedule')


# ## Modify object APIs ###
API_MOD_NODE = str('ModNode')
API_MOD_NODE_CONN = str('ModNodeConn')
API_MOD_RSC = str('ModRsc')
API_TOGGLE_DISK = str('ToggleDisk')
API_MOD_RSC_CONN = str('ModRscConn')
API_MOD_RSC_DFN = str('ModRscDfn')
API_MOD_NET_IF = str('ModNetIf')
API_MOD_STOR_POOL = str('ModStorPool')
API_MOD_STOR_POOL_DFN = str('ModStorPoolDfn')
API_MOD_VLM_DFN = str('ModVlmDfn')
API_MOD_VLM = str('ModVlm')
API_MOD_VLM_CONN = str('ModVlmConn')
API_MOD_SNAPSHOT = str('ModSnapshot')
API_MOD_CRYPT_PASS = str('ModCryptPass')
API_ENABLE_DRBD_PROXY = str('EnableDrbdProxy')
API_DISABLE_DRBD_PROXY = str('DisableDrbdProxy')
API_MOD_DRBD_PROXY = str('ModifyDrbdProxy')
API_ROLLBACK_SNAPSHOT = str('RollbackSnapshot')
API_SHIP_SNAPSHOT = str('ShipSnapshot')
API_MOD_KVS = str('ModifyKvs')
API_MOD_RSC_GRP = str('ModifyRscGrp')
API_MOD_VLM_GRP = str('ModifyVlmGrp')
API_ACTIVATE_RSC = str('ActivateRsc')
API_DEACTIVATE_RSC = str('DeactivateRsc')
API_MOD_EXOS_DFLTS = str('ModifyExosDefaults')
API_MOD_EXOS_ENCLOSURE = str('ModExosEnclosure')
API_MOD_SCHEDULE = str('ModSchedule')


# ## Delete object APIs ###
API_DEL_NODE = str('DelNode')
API_DEL_RSC = str('DelRsc')
API_DEL_RSC_DFN = str('DelRscDfn')
API_DEL_NET_IF = str('DelNetIf')
API_DEL_VLM_DFN = str('DelVlmDfn')
API_DEL_STOR_POOL_DFN = str('DelStorPoolDfn')
API_DEL_STOR_POOL = str('DelStorPool')
API_DEL_NODE_CONN = str('DelNodeConn')
API_DEL_RSC_CONN = str('DelRscConn')
API_DEL_VLM_CONN = str('DelVlmConn')
API_DEL_SNAPSHOT = str('DelSnapshot')
API_DEL_KVS = str('DelKvs')
API_DEL_RSC_GRP = str('DelRscGrp')
API_DEL_VLM_GRP = str('DelVlmGrp')
API_DEL_BACKUP = str('DelBackup')
API_LOST_NODE = str('LostNode')
API_LOST_STOR_POOL = str('LostStorPool')
API_DEL_EXOS_ENCLOSURE = str('DelExosEnclosure')
API_DEL_SCHEDULE = str('DelSchedule')


# ## Authentication APIs ###
API_SIGN_IN = str('SignIn')
API_VERSION = str('Version')


# ## Debug APIs ###
API_CRT_DBG_CNSL = str('CrtDbgCnsl')
API_DSTR_DBG_CNSL = str('DstrDbgCnsl')


# ## Command APIs ###
API_CONTROL_CTRL = str('ControlCtrl')
API_CMD_SHUTDOWN = str('Shutdown')
API_NODE_RECONNECT = str('NodeReconnect')
API_NODE_RESTORE = str('NodeRestore')
API_NODE_EVICT = str('NodeEvict')
API_NODE_EVACUATE = str('NodeEvacuate')


# ## List object APIs ###
API_LST_NODE = str('LstNode')
API_LST_RSC = str('LstRsc')
API_LST_RSC_DFN = str('LstRscDfn')
API_LST_NET_IF = str('LstNetIf')
API_LST_VLM_DFN = str('LstVlmDfn')
API_LST_VLM = str('LstVlm')
API_LST_SNAPSHOT_DFN = str('LstSnapshotDfn')
API_LST_STOR_POOL = str('LstStorPool')
API_LST_STOR_POOL_DFN = str('LstStorPoolDfn')
API_LST_ERROR_REPORTS = str('LstErrorReports')
API_REQ_ERROR_REPORTS = str('ReqErrorReports')
API_DEL_ERROR_REPORT = str('DelErrorReport')
API_DEL_ERROR_REPORTS = str('DelErrorReports')
API_REQ_SOS_REPORT = str('ReqSosReport')
API_REQ_RSC_CONN_LIST = str('ReqRscConnList')
API_LST_RSC_CONN = str('LstRscConn')
API_LST_KVS = str('LstKvs')
API_LST_RSC_GRP = str('LstRscGrp')
API_LST_VLM_GRP = str('LstVlmGrp')
API_LST_PHYS_STOR = str('LstPhysicalStorage')
API_LST_SNAPSHOT_SHIPPINGS = str('LstSnapShips')
API_LST_BACKUPS = str('LstBackups')
API_BACKUP_INFO = str('BackupInfo')
API_LST_PROPS_INFO = str('LstPropsInfo')
API_LST_REMOTE = str('LstRemote')
API_LST_EXOS_DFLTS = str('LstExosDefaults')
API_LST_EXOS_ENCLOSURES = str('LstExosEnclosures')
API_EXOS_ENCLOSURE_EVENTS = str('ExosEvents')
API_EXOS_EXEC = str('ExosExec')
API_EXOS_MAP = str('ExosMap')
API_LST_EXT_FILES = str('LstExtFiles')
API_NODE_STATS = str('NodeStats')
API_RSC_GRP_STATS = str('RscGrpStats')
API_RSC_DFN_STATS = str('RscDfnStats')
API_RSC_STATS = str('RscStats')
API_STOR_POOL_STATS = str('StorPoolStats')
API_ERR_REPORT_STATS = str('ErrReportStats')
API_LST_SCHEDULE = str('LstSchedule')


# ## Query APIs ###
API_QRY_MAX_VLM_SIZE = str('QryMaxVlmSize')
API_RSP_MAX_VLM_SIZE = str('RspMaxVlmSize')
API_RSCDFN_SYNC_STATUS = str('RscDfnSyncStatus')


# ## Event APIs ###
API_CRT_WATCH = str('CrtWatch')
API_DEL_WATCH = str('DelWatch')
API_EVENT = str('Event')
API_RPT_SPC = str('RptSpc')
API_PING = str('Ping')
API_PONG = str('Pong')
API_MOD_INF = str('ModInf')
API_VSN_INF = str('VsnInf')
API_SET_CTRL_PROP = str('SetCtrlProp')
API_DEL_CTRL_PROP = str('DelCtrlProp')
API_LST_CTRL_PROPS = str('LstCtrlProps')


# ## Encryption APIs ###
API_ENTER_CRYPT_PASS = str('EnterCryptPass')


# ## External files APIs ###
API_SET_EXT_FILE = str('SetExtFile')
API_DEL_EXT_FILE = str('DeleteExtFile')
API_DEPLOY_EXT_FILE = str('DeployExtFile')
API_UNDEPLOY_EXT_FILE = str('UndeployExtFile')


# ## DRBD property keys ###
KEY_UUID = str('UUID')
KEY_DRBD_CURRENT_GI = str('DrbdCurrentGi')
KEY_DMSTATS = str('DMStats')
KEY_DRBD_AUTO_QUORUM = str('auto-quorum')
KEY_DRBD_AUTO_ADD_QUORUM_TIEBREAKER = str('auto-add-quorum-tiebreaker')
KEY_MINOR_NR_AUTO_RANGE = str('MinorNrAutoRange')
KEY_DRBD_AUTO_DISKFUL = str('auto-diskful')
KEY_DRBD_AUTO_DISKFUL_ALLOW_CLEANUP = str('auto-diskful-allow-cleanup')
KEY_DRBD_DISABLE_AUTO_VERIFY_ALGO = str('auto-verify-algo-disable')
KEY_DRBD_AUTO_VERIFY_ALGO_ALLOWED_USER = str('auto-verify-algo-allowed-user-list')


# ## Node property keys ###
KEY_NODE = str('Node')
KEY_1ST_NODE = str('FirstNode')
KEY_2ND_NODE = str('SecondNode')
KEY_CUR_STLT_CONN_NAME = str('CurStltConnName')


# ## Resource property keys ###
KEY_RSC_DFN = str('RscDfn')
KEY_RSC_GRP = str('RscGrp')
KEY_TCP_PORT_AUTO_RANGE = str('TcpPortAutoRange')
KEY_PEER_SLOTS_NEW_RESOURCE = str('PeerSlotsNewResource')
KEY_PEER_SLOTS = str('PeerSlots')
KEY_RSC_ROLLBACK_TARGET = str('RollbackTarget')
KEY_RSC_MIGRATE_FROM = str('MigrateFrom')
KEY_RSC_ALLOW_MIXING_DEVICE_KIND = str('AllowMixingStoragePoolDriver')
KEY_RSC_DISKFUL_BY = str('DiskfulBy')


# ## Volume property keys ###
KEY_VLM_GRP = str('VlmGrp')
KEY_VLM_NR = str('VlmNr')
KEY_VLM_RESTORE_FROM_RESOURCE = str('RestoreFromResource')
KEY_VLM_RESTORE_FROM_SNAPSHOT = str('RestoreFromSnapshot')


# ## ldap property keys ###
KEY_SEARCH_DOMAIN = str('SearchDomain')


# ## nvme property keys ###
KEY_TR_TYPE = str('TRType')


# ## Snapshot property keys ###
KEY_SNAPSHOT = str('Snapshot')
KEY_SNAPSHOT_DFN_SEQUENCE_NUMBER = str('SequenceNumber')


# ## Network Interface property keys ###
KEY_PORT = str('Port')
KEY_DISABLE_HTTP_METRICS = str('disable-http-metrics')


# ## Writecache property keys ###
KEY_WRITECACHE_BLOCKSIZE = str('Blocksize')
KEY_WRITECACHE_POOL_NAME = str('PoolName')
KEY_WRITECACHE_SIZE = str('Size')
KEY_WRITECACHE_OPTION_HIGH_WATERMARK = str('HighWatermark')
KEY_WRITECACHE_OPTION_LOW_WATERMARK = str('LowWatermark')
KEY_WRITECACHE_OPTION_START_SECTOR = str('StartSector')
KEY_WRITECACHE_OPTION_WRITEBACK_JOBS = str('WritebackJobs')
KEY_WRITECACHE_OPTION_AUTOCOMMIT_BLOCKS = str('AutocommitBlocks')
KEY_WRITECACHE_OPTION_AUTOCOMMIT_TIME = str('AutocommitTime')
KEY_WRITECACHE_OPTION_FUA = str('Fua')
KEY_WRITECACHE_OPTION_ADDITIONAL = str('Additional')


# ## Cache property keys ###
KEY_CACHE_OPERATING_MODE = str('OpMode')
KEY_CACHE_META_POOL_NAME = str('MetaPool')
KEY_CACHE_CACHE_POOL_NAME = str('CachePool')
KEY_CACHE_META_SIZE = str('Metasize')
KEY_CACHE_CACHE_SIZE = str('Cachesize')
KEY_CACHE_BLOCK_SIZE = str('Blocksize')
KEY_CACHE_POLICY = str('Policy')
KEY_UPDATE_CACHE_INTERVAL = str('UpdateCacheInterval')


# ## BCache property keys ###
KEY_BCACHE_POOL_NAME = str('PoolName')
KEY_BCACHE_SIZE = str('Size')
KEY_BCACHE_BLOCKSIZE = str('Blocksize')
KEY_BCACHE_BUCKETSIZE = str('Bucketsize')
KEY_BCACHE_DATA_OFFSET = str('DataOffset')
KEY_BCACHE_WRITEBACK = str('Writeback')
KEY_BCACHE_DISCARD = str('Discard')
KEY_BCACHE_CACHE_REPLACEMENT_POLICY = str('CacheReplacementPolicy')


# ## Autoplace property keys ###
KEY_AUTOPLACE_STRAT_WEIGHT_MAX_FREESPACE = str('MaxFreeSpace')
KEY_AUTOPLACE_STRAT_WEIGHT_MIN_RESERVED_SPACE = str('MinReservedSpace')
KEY_AUTOPLACE_STRAT_WEIGHT_MIN_RSC_COUNT = str('MinRscCount')
KEY_AUTOPLACE_PRE_SELECT_FILE_NAME = str('PreSelectScript')
KEY_AUTOPLACE_PRE_SELECT_SCRIPT_TIMEOUT = str('PreSelectScriptTimeout')
KEY_AUTOPLACE_MAX_THROUGHPUT = str('MaxThroughput')
KEY_SITE = str('Site')
KEY_AUTOPLACE_ALLOW_TARGET = str('AutoplaceTarget')


# ## Auto-Evict property keys ###
KEY_AUTO_EVICT_MIN_REPLICA_COUNT = str('AutoEvictMinReplicaCount')
KEY_AUTO_EVICT_AFTER_TIME = str('AutoEvictAfterTime')
KEY_AUTO_EVICT_MAX_DISCONNECTED_NODES = str('AutoEvictMaxDisconnectedNodes')
KEY_AUTO_EVICT_ALLOW_EVICTION = str('AutoEvictAllowEviction')


# ## Snapshot shipping property keys ###
KEY_SNAPSHOT_SHIPPING_PREFIX = str('SnapshotShippingPrefix')
KEY_TARGET_NODE = str('TargetNode')
KEY_SOURCE_NODE = str('SourceNode')
KEY_RUN_EVERY = str('RunEvery')
KEY_AUTO_SNAPSHOT_PREFIX = str('Prefix')
KEY_KEEP = str('Keep')
KEY_AUTO_SNAPSHOT_NEXT_ID = str('NextAutoId')
KEY_TCP_PORT_RANGE = str('TcpPortRange')


# ## Backup shipping property keys ###
KEY_BACKUP_TIMEOUT = str('BackupTimeout')
KEY_BACKUP_S3_SUFFIX = str('S3KeySuffix')
KEY_REMOTE = str('Remote')


# ## Clone property keys ###
KEY_USE_ZFS_CLONE = str('UseZFSClone')


# ## NetCom namespaces ###
KEY_NETCOM_ENABLED = str('Enabled')
KEY_NETCOM_BIND_ADDRESS = str('BindAddress')
KEY_NETCOM_KEY_PASSWD = str('KeyPasswd')
KEY_NETCOM_KEY_STORE = str('KeyStore')
KEY_NETCOM_KEY_STORE_PASSWD = str('KeyStorePasswd')
KEY_NETCOM_PORT = str('Port')
KEY_NETCOM_SSL_PROTOCOL = str('SslProtocol')
KEY_NETCOM_TRUST_STORE = str('TrustStore')
KEY_NETCOM_TRUST_STORE_PASSWD = str('TrustStorePasswd')
KEY_NETCOM_TYPE = str('Type')


# ## Property namespaces ###
NAMESPC_NETCOM = str('NetCom')
NAMESPC_DFLT = str('Default')
NAMESPC_LOGGING = str('Logging')
NAMESPC_ALLOC = str('Allocation')
NAMESPC_NETIF = str('NetIf')
NAMESPC_STLT = str('Satellite')
NAMESPC_NODE = str('Node')
NAMESPC_STORAGE_DRIVER = str('StorDriver')
NAMESPC_DRBD_PROXY = str('DrbdProxy')
NAMESPC_AUXILIARY = str('Aux')
NAMESPC_DRBD_OPTIONS = str('DrbdOptions')
NAMESPC_DRBD_NET_OPTIONS = str('DrbdOptions/Net')
NAMESPC_DRBD_DISK_OPTIONS = str('DrbdOptions/Disk')
NAMESPC_DRBD_RESOURCE_OPTIONS = str('DrbdOptions/Resource')
NAMESPC_DRBD_PEER_DEVICE_OPTIONS = str('DrbdOptions/PeerDevice')
NAMESPC_DRBD_PROXY_OPTIONS = str('DrbdOptions/Proxy')
NAMESPC_DRBD_PROXY_COMPRESSION_OPTIONS = str('DrbdOptions/ProxyCompression')
NAMESPC_DRBD_HANDLER_OPTIONS = str('DrbdOptions/Handlers')
NAMESPC_CONNECTION_PATHS = str('Paths')
NAMESPC_REST = str('REST')
NAMESPC_FILESYSTEM = str('FileSystem')
NAMESPC_NVME = str('NVMe')
NAMESPC_SYS_FS = str('sys/fs')
NAMESPC_WRITECACHE = str('Writecache')
NAMESPC_WRITECACHE_OPTIONS = str('Writecache/Options')
NAMESPC_CACHE = str('Cache')
NAMESPC_CACHE_FEATURES = str('Cache/Features')
NAMESPC_CACHE_POLICY_ARGS = str('Cache/Policy')
NAMESPC_BCACHE = str('BCache')
NAMESPC_AUTOPLACER = str('Autoplacer')
NAMESPC_AUTOPLACER_WEIGHTS = str('Autoplacer/Weights')
NAMESPC_SNAPSHOT_SHIPPING = str('SnapshotShipping')
NAMESPC_AUTO_SNAPSHOT = str('AutoSnapshot')
NAMESPC_STLT_DEV_SYMLINKS = str('Satellite/Device/Symlinks')
NAMESPC_EXOS = str('StorDriver/Exos')
NAMESPC_BACKUP_SHIPPING = str('BackupShipping')
NAMESPC_CLUSTER = str('Cluster')
NAMESPC_CLUSTER_REMOTE = str('Cluster/Remote')
NAMESPC_SED = str('SED')
NAMESPC_EBS = str('EBS')
NAMESPC_TAGS = str('Tags')


# ## Storage pool property keys ###
KEY_STOR_POOL_DFN = str('StorPoolDfn')
KEY_STOR_POOL_NAME = str('StorPoolName')
KEY_STOR_POOL_DRBD_META_NAME = str('StorPoolNameDrbdMeta')
KEY_STOR_POOL_VOLUME_GROUP = str('LvmVg')
KEY_STOR_POOL_LVCREATE_TYPE = str('LvcreateType')
KEY_STOR_POOL_LVCREATE_OPTIONS = str('LvcreateOptions')
KEY_STOR_POOL_THIN_POOL = str('ThinPool')
KEY_STOR_POOL_ZPOOL = str('ZPool')
KEY_STOR_POOL_ZPOOLTHIN = str('ZPoolThin')
KEY_STOR_POOL_ZFS_CREATE_OPTIONS = str('ZfscreateOptions')
KEY_STOR_POOL_FILE_DIRECTORY = str('FileDir')
KEY_STOR_POOL_PREF_NIC = str('PrefNic')
KEY_STOR_POOL_CRYPT_PASSWD = str('CryptPasswd')
KEY_STOR_POOL_OVERRIDE_VLM_ID = str('OverrideVlmId')
KEY_STOR_POOL_DFN_MAX_OVERSUBSCRIPTION_RATIO = str('MaxOversubscriptionRatio')
KEY_STOR_POOL_WAIT_TIMEOUT_AFTER_CREATE = str('WaitTimeoutAfterCreate')
KEY_STOR_POOL_OPENFLEX_API_HOST = str('Openflex/ApiHost')
KEY_STOR_POOL_OPENFLEX_API_PORT = str('Openflex/ApiPort')
KEY_STOR_POOL_OPENFLEX_STOR_DEV = str('Openflex/StorDev')
KEY_STOR_POOL_OPENFLEX_STOR_DEV_HOST = str('Openflex/StorDevHost')
KEY_STOR_POOL_OPENFLEX_STOR_POOL = str('Openflex/StorPool')
KEY_STOR_POOL_OPENFLEX_USER_NAME = str('Openflex/UserName')
KEY_STOR_POOL_OPENFLEX_USER_PW = str('Openflex/UserPassword')
KEY_STOR_POOL_OPENFLEX_JOB_WAIT_MAX_COUNT = str('Openflex/JobWaitMaxCount')
KEY_STOR_POOL_OPENFLEX_JOB_WAIT_DELAY = str('Openflex/JobWaitDelay')
KEY_OF_TARGET_PORT_AUTO_RANGE = str('OpenflexTargetPortAutoRange')
KEY_STOR_POOL_EXOS_API_IP = str('IP')
KEY_STOR_POOL_EXOS_API_IP_ENV = str('IPEnv')
KEY_STOR_POOL_EXOS_API_PORT = str('Port')
KEY_STOR_POOL_EXOS_API_USER = str('Username')
KEY_STOR_POOL_EXOS_API_USER_ENV = str('UsernameEnv')
KEY_STOR_POOL_EXOS_API_PASSWORD = str('Password')
KEY_STOR_POOL_EXOS_API_PASSWORD_ENV = str('PasswordEnv')
KEY_STOR_POOL_EXOS_VLM_TYPE = str('VolumeType')
KEY_STOR_POOL_EXOS_CREATE_VOLUME_OPTIONS = str('CreateVolumeOptions')
KEY_STOR_POOL_EXOS_ENCLOSURE = str('Enclosure')
KEY_STOR_POOL_EXOS_POOL_SN = str('PoolSN')
KEY_STOR_POOL_REMOTE_SPDK_API_HOST = str('RemoteSpdk/ApiHost')
KEY_STOR_POOL_REMOTE_SPDK_API_PORT = str('RemoteSpdk/ApiPort')
KEY_STOR_POOL_REMOTE_SPDK_API_USER_NAME = str('RemoteSpdk/UserName')
KEY_STOR_POOL_REMOTE_SPDK_API_USER_PW = str('RemoteSpdk/UserPassword')
KEY_STOR_POOL_REMOTE_SPDK_API_USER_NAME_ENV = str('RemoteSpdk/UserNameEnv')
KEY_STOR_POOL_REMOTE_SPDK_API_USER_PW_ENV = str('RemoteSpdk/UserPasswordEnv')
KEY_EBS_VOLUME_TYPE = str('EBS/VolumeType')
KEY_PREF_NIC = str('PrefNic')


# ## Storage pool traits keys ###
KEY_STOR_POOL_SUPPORTS_SNAPSHOTS = str('SupportsSnapshots')
KEY_STOR_POOL_PROVISIONING = str('Provisioning')
# Unit of smallest allocation. The size in KiB as a decimal number.
KEY_STOR_POOL_ALLOCATION_UNIT = str('AllocationUnit')


# ## Storage pool traits values ###
VAL_STOR_POOL_PROVISIONING_FAT = str('Fat')
VAL_STOR_POOL_PROVISIONING_THIN = str('Thin')
VAL_STOR_POOL_DRBD_META_INTERNAL = str('.internal')
VAL_STOR_POOL_SPACE_NOT_FOUND = long(-1)
VAL_STOR_POOL_SPACE_ENOUGH = long(9223372036854775807)


# ## DRBD Proxy keys (other than 'options') ###
KEY_DRBD_PROXY_COMPRESSION_TYPE = str('CompressionType')
KEY_DRBD_PROXY_AUTO_ENABLE = str('AutoEnable')


# ## File system property keys ###
KEY_FS_TYPE = str('Type')
KEY_FS_MKFSPARAMETERS = str('MkfsParams')
VAL_FS_TYPE_EXT4 = str('ext4')
VAL_FS_TYPE_XFS = str('xfs')


# ## sys/fs property keys ###
KEY_SYS_FS_BLKIO_THROTTLE_READ = str('blkio_throttle_read')
KEY_SYS_FS_BLKIO_THROTTLE_WRITE = str('blkio_throttle_write')
KEY_SYS_FS_BLKIO_THROTTLE_READ_IOPS = str('blkio_throttle_read_iops')
KEY_SYS_FS_BLKIO_THROTTLE_WRITE_IOPS = str('blkio_throttle_write_iops')


# ## Property values ###
VAL_NETCOM_TYPE_SSL = str('SSL')
VAL_NETCOM_TYPE_PLAIN = str('Plain')
VAL_SSL_PROTO_TLSv1 = str('TLSv1')


# ## DRBD related property values ###
VAL_DRBD_PROXY_COMPRESSION_NONE = str('none')
VAL_DRBD_PROXY_COMPRESSION_ZSTD = str('zstd')
VAL_DRBD_PROXY_COMPRESSION_ZLIB = str('zlib')
VAL_DRBD_PROXY_COMPRESSION_LZMA = str('lzma')
VAL_DRBD_PROXY_COMPRESSION_LZ4 = str('lz4')
VAL_DRBD_AUTO_QUORUM_DISABLED = str('disabled')
VAL_DRBD_AUTO_QUORUM_IO_ERROR = str('io-error')
VAL_DRBD_AUTO_QUORUM_SUSPEND_IO = str('suspend-io')


# ## Node Type values ###
VAL_NODE_TYPE_CTRL = str('Controller')
VAL_NODE_TYPE_STLT = str('Satellite')
VAL_NODE_TYPE_CMBD = str('Combined')
VAL_NODE_TYPE_AUX = str('Auxiliary')
VAL_NODE_TYPE_OPENFLEX_TARGET = str('Openflex_Target')
VAL_NODE_TYPE_REMOTE_SPDK = str('Remote_Spdk')
VAL_NODE_TYPE_EXOS_TARGET = str('Exos_Target')


# ## Writecache option values ###
VAL_WRITECACHE_FUA_ON = str('On')
VAL_WRITECACHE_FUA_OFF = str('Off')


# ## Net interface Type values ###
VAL_NETIF_TYPE_IP = str('IP')
VAL_NETIF_TYPE_RDMA = str('RDMA')
VAL_NETIF_TYPE_ROCE = str('RoCE')


# ## Authentication keys ###
KEY_SEC_IDENTITY = str('SecIdentity')
KEY_SEC_ROLE = str('SecRole')
KEY_SEC_TYPE = str('SecType')
KEY_SEC_DOMAIN = str('SecDomain')
KEY_SEC_PASSWORD = str('SecPassword')
KEY_POOL_NAME = str('PoolName')


# ## External commands keys ###
KEY_EXT_CMD_WAIT_TO = str('ExtCmdWaitTimeout')


# ## External files keys ###
KEY_EXT_FILE = str('ExtFile')


# ## Default ports ###
DFLT_CTRL_PORT_SSL = int(3371)
DFLT_CTRL_PORT_PLAIN = int(3370)
DFLT_STLT_PORT_SSL = int(3367)
DFLT_STLT_PORT_PLAIN = int(3366)


# ## Boolean values ###
VAL_TRUE = str('True')
VAL_FALSE = str('False')
VAL_YES = str('Yes')
VAL_NO = str('No')


# ## Snapshot-shipping values ###
VAL_SNAP_SHIP_NAME = str('SnapshotShipping')


class SnapshotShipStatus(Enum):
    RUNNING = "Running"
    COMPLETE = "Complete"


# ## Flag string values ###
FLAG_CLEAN = str('CLEAN')
FLAG_EVICTED = str('EVICTED')
FLAG_EVACUATE = str('EVACUATE')
FLAG_DELETE = str('DELETE')
FLAG_DISKLESS = str('DISKLESS')
FLAG_QIGNORE = str('QIGNORE')
FLAG_ENCRYPTED = str('ENCRYPTED')
FLAG_GROSS_SIZE = str('GROSS_SIZE')
FLAG_SUCCESSFUL = str('SUCCESSFUL')
FLAG_SHIPPING = str('SHIPPING')
FLAG_SHIPPING_CLEANUP = str('SHIPPING_CLEANUP')
FLAG_SHIPPING_ABORT = str('SHIPPING_ABORT')
FLAG_SHIPPED = str('SHIPPED')
FLAG_AUTO_SNAPSHOT = str('AUTO_SNAPSHOT')
FLAG_BACKUP = str('BACKUP')
FLAG_RESTORE_BACKUP_ON_SUCCESS = str('RESTORE_BACKUP_ON_SUCCESS')
FLAG_BACKUP_TARGET = str('BACKUP_TARGET')
FLAG_BACKUP_SOURCE = str('BACKUP_SOURCE')
FLAG_FAILED_DEPLOYMENT = str('FAILED_DEPLOYMENT')
FLAG_FAILED_DISCONNECT = str('FAILED_DISCONNECT')
FLAG_RESIZE = str('RESIZE')
FLAG_DISK_ADDING = str('DISK_ADDING')
FLAG_DISK_ADD_REQUESTED = str('DISK_ADD_REQUESTED')
FLAG_DISK_REMOVING = str('DISK_REMOVING')
FLAG_DISK_REMOVE_REQUESTED = str('DISK_REMOVE_REQUESTED')
FLAG_TIE_BREAKER = str('TIE_BREAKER')
FLAG_DRBD_DISKLESS = str('DRBD_DISKLESS')
FLAG_NVME_INITIATOR = str('NVME_INITIATOR')
FLAG_RSC_INACTIVE = str('INACTIVE')
FLAG_EBS_INITIATOR = str('EBS_INITIATOR')


# ## Device layer kinds ###


class DeviceLayerKind(Enum):
    DRBD = "DRBD"
    LUKS = "LUKS"
    STORAGE = "STORAGE"
    NVME = "NVME"
    OPENFLEX = "OPENFLEX"
    EXOS = "EXOS"
    WRITECACHE = "WRITECACHE"
    CACHE = "CACHE"
    BCACHE = "BCACHE"


# ## Satellite connection statuses ###


class ConnectionStatus(Enum):
    OFFLINE = 0
    CONNECTED = 1
    ONLINE = 2
    VERSION_MISMATCH = 3
    FULL_SYNC_FAILED = 4
    AUTHENTICATION_ERROR = 5
    UNKNOWN = 6
    HOSTNAME_MISMATCH = 7
    OTHER_CONTROLLER = 8
    AUTHENTICATED = 9
    NO_STLT_CONN = 10


# ## Clone states ###


class CloneStatus(Enum):
    FAILED = "FAILED"
    CLONING = "CLONING"
    COMPLETE = "COMPLETE"


# ## Default names ###
DEFAULT_NETIF = str('default')
DFLT_SNAPSHOT_SHIPPING_PREFIX = str('ship')


# ## Default values ###
DFLT_AUTO_SNAPSHOT_KEEP = str('10')
DFLT_SHIPPED_SNAPSHOT_KEEP = str('10')


if __name__ == "__main__":
    MAP_MASK = {
        'MASK_BITS_TYPE': MASK_BITS_TYPE,
        'MASK_ERROR': MASK_ERROR,
        'MASK_WARN': MASK_WARN,
        'MASK_INFO': MASK_INFO,
        'MASK_SUCCESS': MASK_SUCCESS,
        'MASK_BITS_OP': MASK_BITS_OP,
        'MASK_CRT': MASK_CRT,
        'MASK_MOD': MASK_MOD,
        'MASK_DEL': MASK_DEL,
        'MASK_BITS_OBJ': MASK_BITS_OBJ,
        'MASK_SCHEDULE': MASK_SCHEDULE,
        'MASK_EXT_FILES': MASK_EXT_FILES,
        'MASK_PHYSICAL_DEVICE': MASK_PHYSICAL_DEVICE,
        'MASK_VLM_GRP': MASK_VLM_GRP,
        'MASK_RSC_GRP': MASK_RSC_GRP,
        'MASK_KVS': MASK_KVS,
        'MASK_NODE': MASK_NODE,
        'MASK_RSC_DFN': MASK_RSC_DFN,
        'MASK_RSC': MASK_RSC,
        'MASK_VLM_DFN': MASK_VLM_DFN,
        'MASK_VLM': MASK_VLM,
        'MASK_NODE_CONN': MASK_NODE_CONN,
        'MASK_RSC_CONN': MASK_RSC_CONN,
        'MASK_VLM_CONN': MASK_VLM_CONN,
        'MASK_NET_IF': MASK_NET_IF,
        'MASK_STOR_POOL_DFN': MASK_STOR_POOL_DFN,
        'MASK_STOR_POOL': MASK_STOR_POOL,
        'MASK_CTRL_CONF': MASK_CTRL_CONF,
        'MASK_SNAPSHOT': MASK_SNAPSHOT,
        'MASK_BACKUP': MASK_BACKUP,
        'MASK_REMOTE': MASK_REMOTE,
        'MASK_BITS_CODE': MASK_BITS_CODE,
        'CREATED': CREATED,
        'DELETED': DELETED,
        'MODIFIED': MODIFIED,
        'PASSPHRASE_ACCEPTED': PASSPHRASE_ACCEPTED,
        'FAIL_SQL': FAIL_SQL,
        'FAIL_SQL_ROLLBACK': FAIL_SQL_ROLLBACK,
        'FAIL_INVLD_NODE_NAME': FAIL_INVLD_NODE_NAME,
        'FAIL_INVLD_NODE_TYPE': FAIL_INVLD_NODE_TYPE,
        'FAIL_INVLD_RSC_NAME': FAIL_INVLD_RSC_NAME,
        'FAIL_INVLD_RSC_PORT': FAIL_INVLD_RSC_PORT,
        'FAIL_INVLD_NODE_ID': FAIL_INVLD_NODE_ID,
        'FAIL_INVLD_VLM_NR': FAIL_INVLD_VLM_NR,
        'FAIL_INVLD_VLM_SIZE': FAIL_INVLD_VLM_SIZE,
        'FAIL_INVLD_MINOR_NR': FAIL_INVLD_MINOR_NR,
        'FAIL_INVLD_STOR_POOL_NAME': FAIL_INVLD_STOR_POOL_NAME,
        'FAIL_INVLD_NET_NAME': FAIL_INVLD_NET_NAME,
        'FAIL_INVLD_NET_ADDR': FAIL_INVLD_NET_ADDR,
        'FAIL_INVLD_NET_PORT': FAIL_INVLD_NET_PORT,
        'FAIL_INVLD_NET_TYPE': FAIL_INVLD_NET_TYPE,
        'FAIL_INVLD_PROP': FAIL_INVLD_PROP,
        'FAIL_INVLD_TRANSPORT_TYPE': FAIL_INVLD_TRANSPORT_TYPE,
        'FAIL_INVLD_TCP_PORT': FAIL_INVLD_TCP_PORT,
        'FAIL_INVLD_CRYPT_PASSPHRASE': FAIL_INVLD_CRYPT_PASSPHRASE,
        'FAIL_INVLD_ENCRYPT_TYPE': FAIL_INVLD_ENCRYPT_TYPE,
        'FAIL_INVLD_SNAPSHOT_NAME': FAIL_INVLD_SNAPSHOT_NAME,
        'FAIL_INVLD_PLACE_COUNT': FAIL_INVLD_PLACE_COUNT,
        'FAIL_INVLD_FREE_SPACE_MGR_NAME': FAIL_INVLD_FREE_SPACE_MGR_NAME,
        'FAIL_INVLD_STOR_DRIVER': FAIL_INVLD_STOR_DRIVER,
        'FAIL_INVLD_DRBD_PROXY_COMPRESSION_TYPE': FAIL_INVLD_DRBD_PROXY_COMPRESSION_TYPE,
        'FAIL_INVLD_KVS_NAME': FAIL_INVLD_KVS_NAME,
        'FAIL_INVLD_LAYER_KIND': FAIL_INVLD_LAYER_KIND,
        'FAIL_INVLD_LAYER_STACK': FAIL_INVLD_LAYER_STACK,
        'FAIL_INVLD_EXT_NAME': FAIL_INVLD_EXT_NAME,
        'FAIL_INVLD_PROVIDER': FAIL_INVLD_PROVIDER,
        'FAIL_INVLD_VLM_SIZES': FAIL_INVLD_VLM_SIZES,
        'FAIL_INVLD_VLM_COUNT': FAIL_INVLD_VLM_COUNT,
        'FAIL_INVLD_CONF': FAIL_INVLD_CONF,
        'FAIL_INVLD_SNAPSHOT_SHIPPING_SOURCE': FAIL_INVLD_SNAPSHOT_SHIPPING_SOURCE,
        'FAIL_INVLD_SNAPSHOT_SHIPPING_TARGET': FAIL_INVLD_SNAPSHOT_SHIPPING_TARGET,
        'FAIL_NODE_HAS_USED_RSC': FAIL_NODE_HAS_USED_RSC,
        'FAIL_INVLD_REQUEST': FAIL_INVLD_REQUEST,
        'FAIL_INVLD_EXT_FILE_NAME': FAIL_INVLD_EXT_FILE_NAME,
        'FAIL_INVLD_EXT_FILE': FAIL_INVLD_EXT_FILE,
        'FAIL_INVLD_RSC_GRP_NAME': FAIL_INVLD_RSC_GRP_NAME,
        'FAIL_INVLD_BACKUP_CONFIG': FAIL_INVLD_BACKUP_CONFIG,
        'FAIL_INVLD_REMOTE_NAME': FAIL_INVLD_REMOTE_NAME,
        'FAIL_INVLD_TIME_PARAM': FAIL_INVLD_TIME_PARAM,
        'FAIL_INVLD_SCHEDULE_NAME': FAIL_INVLD_SCHEDULE_NAME,
        'FAIL_NOT_FOUND_NODE': FAIL_NOT_FOUND_NODE,
        'FAIL_NOT_FOUND_RSC_DFN': FAIL_NOT_FOUND_RSC_DFN,
        'FAIL_NOT_FOUND_RSC': FAIL_NOT_FOUND_RSC,
        'FAIL_NOT_FOUND_VLM_DFN': FAIL_NOT_FOUND_VLM_DFN,
        'FAIL_NOT_FOUND_VLM': FAIL_NOT_FOUND_VLM,
        'FAIL_NOT_FOUND_NET_IF': FAIL_NOT_FOUND_NET_IF,
        'FAIL_NOT_FOUND_NODE_CONN': FAIL_NOT_FOUND_NODE_CONN,
        'FAIL_NOT_FOUND_RSC_CONN': FAIL_NOT_FOUND_RSC_CONN,
        'FAIL_NOT_FOUND_VLM_CONN': FAIL_NOT_FOUND_VLM_CONN,
        'FAIL_NOT_FOUND_STOR_POOL_DFN': FAIL_NOT_FOUND_STOR_POOL_DFN,
        'FAIL_NOT_FOUND_STOR_POOL': FAIL_NOT_FOUND_STOR_POOL,
        'FAIL_NOT_FOUND_DFLT_STOR_POOL': FAIL_NOT_FOUND_DFLT_STOR_POOL,
        'FAIL_NOT_FOUND_CRYPT_KEY': FAIL_NOT_FOUND_CRYPT_KEY,
        'FAIL_NOT_FOUND_SNAPSHOT_DFN': FAIL_NOT_FOUND_SNAPSHOT_DFN,
        'FAIL_NOT_FOUND_SNAPSHOT_VLM_DFN': FAIL_NOT_FOUND_SNAPSHOT_VLM_DFN,
        'FAIL_NOT_FOUND_SNAPSHOT': FAIL_NOT_FOUND_SNAPSHOT,
        'FAIL_NOT_FOUND_KVS': FAIL_NOT_FOUND_KVS,
        'FAIL_NOT_FOUND_RSC_GRP': FAIL_NOT_FOUND_RSC_GRP,
        'FAIL_NOT_FOUND_VLM_GRP': FAIL_NOT_FOUND_VLM_GRP,
        'FAIL_NOT_FOUND_EXOS_ENCLOSURE': FAIL_NOT_FOUND_EXOS_ENCLOSURE,
        'FAIL_NOT_FOUND_EXT_FILE': FAIL_NOT_FOUND_EXT_FILE,
        'FAIL_NOT_FOUND_REMOTE': FAIL_NOT_FOUND_REMOTE,
        'FAIL_NOT_FOUND_BACKUP': FAIL_NOT_FOUND_BACKUP,
        'FAIL_NOT_FOUND_SCHEDULE': FAIL_NOT_FOUND_SCHEDULE,
        'FAIL_ACC_DENIED_NODE': FAIL_ACC_DENIED_NODE,
        'FAIL_ACC_DENIED_RSC_DFN': FAIL_ACC_DENIED_RSC_DFN,
        'FAIL_ACC_DENIED_RSC': FAIL_ACC_DENIED_RSC,
        'FAIL_ACC_DENIED_VLM_DFN': FAIL_ACC_DENIED_VLM_DFN,
        'FAIL_ACC_DENIED_VLM': FAIL_ACC_DENIED_VLM,
        'FAIL_ACC_DENIED_STOR_POOL_DFN': FAIL_ACC_DENIED_STOR_POOL_DFN,
        'FAIL_ACC_DENIED_STOR_POOL': FAIL_ACC_DENIED_STOR_POOL,
        'FAIL_ACC_DENIED_NODE_CONN': FAIL_ACC_DENIED_NODE_CONN,
        'FAIL_ACC_DENIED_RSC_CONN': FAIL_ACC_DENIED_RSC_CONN,
        'FAIL_ACC_DENIED_VLM_CONN': FAIL_ACC_DENIED_VLM_CONN,
        'FAIL_ACC_DENIED_STLT_CONN': FAIL_ACC_DENIED_STLT_CONN,
        'FAIL_ACC_DENIED_CTRL_CFG': FAIL_ACC_DENIED_CTRL_CFG,
        'FAIL_ACC_DENIED_COMMAND': FAIL_ACC_DENIED_COMMAND,
        'FAIL_ACC_DENIED_WATCH': FAIL_ACC_DENIED_WATCH,
        'FAIL_ACC_DENIED_SNAPSHOT_DFN': FAIL_ACC_DENIED_SNAPSHOT_DFN,
        'FAIL_ACC_DENIED_SNAPSHOT': FAIL_ACC_DENIED_SNAPSHOT,
        'FAIL_ACC_DENIED_SNAPSHOT_VLM_DFN': FAIL_ACC_DENIED_SNAPSHOT_VLM_DFN,
        'FAIL_ACC_DENIED_FREE_SPACE_MGR': FAIL_ACC_DENIED_FREE_SPACE_MGR,
        'FAIL_ACC_DENIED_KVS': FAIL_ACC_DENIED_KVS,
        'FAIL_ACC_DENIED_RSC_GRP': FAIL_ACC_DENIED_RSC_GRP,
        'FAIL_ACC_DENIED_VLM_GRP': FAIL_ACC_DENIED_VLM_GRP,
        'FAIL_ACC_DENIED_SNAP_DFN': FAIL_ACC_DENIED_SNAP_DFN,
        'FAIL_ACC_DENIED_EXT_FILE': FAIL_ACC_DENIED_EXT_FILE,
        'FAIL_ACC_DENIED_REMOTE': FAIL_ACC_DENIED_REMOTE,
        'FAIL_ACC_DENIED_SCHEDULE': FAIL_ACC_DENIED_SCHEDULE,
        'FAIL_EXISTS_NODE': FAIL_EXISTS_NODE,
        'FAIL_EXISTS_RSC_DFN': FAIL_EXISTS_RSC_DFN,
        'FAIL_EXISTS_RSC': FAIL_EXISTS_RSC,
        'FAIL_EXISTS_VLM_DFN': FAIL_EXISTS_VLM_DFN,
        'FAIL_EXISTS_VLM': FAIL_EXISTS_VLM,
        'FAIL_EXISTS_NET_IF': FAIL_EXISTS_NET_IF,
        'FAIL_EXISTS_NODE_CONN': FAIL_EXISTS_NODE_CONN,
        'FAIL_EXISTS_RSC_CONN': FAIL_EXISTS_RSC_CONN,
        'FAIL_EXISTS_VLM_CONN': FAIL_EXISTS_VLM_CONN,
        'FAIL_EXISTS_STOR_POOL_DFN': FAIL_EXISTS_STOR_POOL_DFN,
        'FAIL_EXISTS_STOR_POOL': FAIL_EXISTS_STOR_POOL,
        'FAIL_EXISTS_STLT_CONN': FAIL_EXISTS_STLT_CONN,
        'FAIL_EXISTS_CRYPT_PASSPHRASE': FAIL_EXISTS_CRYPT_PASSPHRASE,
        'FAIL_EXISTS_WATCH': FAIL_EXISTS_WATCH,
        'FAIL_EXISTS_SNAPSHOT_DFN': FAIL_EXISTS_SNAPSHOT_DFN,
        'FAIL_EXISTS_SNAPSHOT': FAIL_EXISTS_SNAPSHOT,
        'FAIL_EXISTS_EXT_NAME': FAIL_EXISTS_EXT_NAME,
        'FAIL_EXISTS_NVME_TARGET_PER_RSC_DFN': FAIL_EXISTS_NVME_TARGET_PER_RSC_DFN,
        'FAIL_EXISTS_NVME_INITIATOR_PER_RSC_DFN': FAIL_EXISTS_NVME_INITIATOR_PER_RSC_DFN,
        'FAIL_LOST_STOR_POOL': FAIL_LOST_STOR_POOL,
        'FAIL_EXISTS_RSC_GRP': FAIL_EXISTS_RSC_GRP,
        'FAIL_EXISTS_VLM_GRP': FAIL_EXISTS_VLM_GRP,
        'FAIL_EXISTS_OPENFLEX_TARGET_PER_RSC_DFN': FAIL_EXISTS_OPENFLEX_TARGET_PER_RSC_DFN,
        'FAIL_EXISTS_SNAPSHOT_SHIPPING': FAIL_EXISTS_SNAPSHOT_SHIPPING,
        'FAIL_EXISTS_EXOS_ENCLOSURE': FAIL_EXISTS_EXOS_ENCLOSURE,
        'FAIL_EXISTS_REMOTE': FAIL_EXISTS_REMOTE,
        'FAIL_EXISTS_SCHEDULE': FAIL_EXISTS_SCHEDULE,
        'FAIL_MISSING_PROPS': FAIL_MISSING_PROPS,
        'FAIL_MISSING_PROPS_NETCOM_TYPE': FAIL_MISSING_PROPS_NETCOM_TYPE,
        'FAIL_MISSING_PROPS_NETCOM_PORT': FAIL_MISSING_PROPS_NETCOM_PORT,
        'FAIL_MISSING_NETCOM': FAIL_MISSING_NETCOM,
        'FAIL_MISSING_PROPS_NETIF_NAME': FAIL_MISSING_PROPS_NETIF_NAME,
        'FAIL_MISSING_STLT_CONN': FAIL_MISSING_STLT_CONN,
        'FAIL_MISSING_EXT_NAME': FAIL_MISSING_EXT_NAME,
        'FAIL_MISSING_NVME_TARGET': FAIL_MISSING_NVME_TARGET,
        'FAIL_NO_STLT_CONN_DEFINED': FAIL_NO_STLT_CONN_DEFINED,
        'FAIL_MISSING_OPENFLEX_TARGET': FAIL_MISSING_OPENFLEX_TARGET,
        'FAIL_MISSING_EBS_TARGET': FAIL_MISSING_EBS_TARGET,
        'FAIL_UUID_NODE': FAIL_UUID_NODE,
        'FAIL_UUID_RSC_DFN': FAIL_UUID_RSC_DFN,
        'FAIL_UUID_RSC': FAIL_UUID_RSC,
        'FAIL_UUID_VLM_DFN': FAIL_UUID_VLM_DFN,
        'FAIL_UUID_VLM': FAIL_UUID_VLM,
        'FAIL_UUID_NET_IF': FAIL_UUID_NET_IF,
        'FAIL_UUID_NODE_CONN': FAIL_UUID_NODE_CONN,
        'FAIL_UUID_RSC_CONN': FAIL_UUID_RSC_CONN,
        'FAIL_UUID_VLM_CONN': FAIL_UUID_VLM_CONN,
        'FAIL_UUID_STOR_POOL_DFN': FAIL_UUID_STOR_POOL_DFN,
        'FAIL_UUID_STOR_POOL': FAIL_UUID_STOR_POOL,
        'FAIL_UUID_KVS': FAIL_UUID_KVS,
        'FAIL_POOL_EXHAUSTED_VLM_NR': FAIL_POOL_EXHAUSTED_VLM_NR,
        'FAIL_POOL_EXHAUSTED_MINOR_NR': FAIL_POOL_EXHAUSTED_MINOR_NR,
        'FAIL_POOL_EXHAUSTED_TCP_PORT': FAIL_POOL_EXHAUSTED_TCP_PORT,
        'FAIL_POOL_EXHAUSTED_NODE_ID': FAIL_POOL_EXHAUSTED_NODE_ID,
        'FAIL_POOL_EXHAUSTED_RSC_LAYER_ID': FAIL_POOL_EXHAUSTED_RSC_LAYER_ID,
        'FAIL_POOL_EXHAUSTED_SPECIAL_SATELLTE_TCP_PORT': FAIL_POOL_EXHAUSTED_SPECIAL_SATELLTE_TCP_PORT,
        'FAIL_POOL_EXHAUSTED_SNAPSHOT_SHIPPING_TCP_PORT': FAIL_POOL_EXHAUSTED_SNAPSHOT_SHIPPING_TCP_PORT,
        'FAIL_POOL_EXHAUSTED_BACKUP_SHIPPING_TCP_PORT': FAIL_POOL_EXHAUSTED_BACKUP_SHIPPING_TCP_PORT,
        'FAIL_EBS_COOLDOWN': FAIL_EBS_COOLDOWN,
        'FAIL_NOT_ALL_UPTODATE': FAIL_NOT_ALL_UPTODATE,
        'FAIL_EVACUATING': FAIL_EVACUATING,
        'FAIL_BACKUP_UNKNOWN_CLUSTER': FAIL_BACKUP_UNKNOWN_CLUSTER,
        'FAIL_DEPENDEND_BACKUP': FAIL_DEPENDEND_BACKUP,
        'FAIL_BACKUP_INCOMPATIBLE_VERSION': FAIL_BACKUP_INCOMPATIBLE_VERSION,
        'FAIL_SNAPSHOT_NOT_UPTODATE': FAIL_SNAPSHOT_NOT_UPTODATE,
        'FAIL_NOT_ENOUGH_FREE_SPACE': FAIL_NOT_ENOUGH_FREE_SPACE,
        'FAIL_ONLY_ONE_ACT_RSC_PER_SHARED_STOR_POOL_ALLOWED': FAIL_ONLY_ONE_ACT_RSC_PER_SHARED_STOR_POOL_ALLOWED,
        'FAIL_CRYPT_INIT': FAIL_CRYPT_INIT,
        'FAIL_SNAPSHOT_SHIPPING_NOT_SUPPORTED': FAIL_SNAPSHOT_SHIPPING_NOT_SUPPORTED,
        'FAIL_SNAPSHOT_SHIPPING_IN_PROGRESS': FAIL_SNAPSHOT_SHIPPING_IN_PROGRESS,
        'FAIL_UNDECIDABLE_AUTOPLACMENT': FAIL_UNDECIDABLE_AUTOPLACMENT,
        'FAIL_PRE_SELECT_SCRIPT_DID_NOT_TERMINATE': FAIL_PRE_SELECT_SCRIPT_DID_NOT_TERMINATE,
        'FAIL_LINSTOR_MANAGED_SATELLITE_DID_NOT_START_PROPERLY': FAIL_LINSTOR_MANAGED_SATELLITE_DID_NOT_START_PROPERLY,
        'FAIL_STLT_DOES_NOT_SUPPORT_LAYER': FAIL_STLT_DOES_NOT_SUPPORT_LAYER,
        'FAIL_STLT_DOES_NOT_SUPPORT_PROVIDER': FAIL_STLT_DOES_NOT_SUPPORT_PROVIDER,
        'FAIL_STOR_POOL_CONFIGURATION_ERROR': FAIL_STOR_POOL_CONFIGURATION_ERROR,
        'FAIL_INSUFFICIENT_REPLICA_COUNT': FAIL_INSUFFICIENT_REPLICA_COUNT,
        'FAIL_RSC_BUSY': FAIL_RSC_BUSY,
        'FAIL_INSUFFICIENT_PEER_SLOTS': FAIL_INSUFFICIENT_PEER_SLOTS,
        'FAIL_SNAPSHOTS_NOT_SUPPORTED': FAIL_SNAPSHOTS_NOT_SUPPORTED,
        'FAIL_NOT_CONNECTED': FAIL_NOT_CONNECTED,
        'FAIL_NOT_ENOUGH_NODES': FAIL_NOT_ENOUGH_NODES,
        'FAIL_IN_USE': FAIL_IN_USE,
        'FAIL_UNKNOWN_ERROR': FAIL_UNKNOWN_ERROR,
        'FAIL_IMPL_ERROR': FAIL_IMPL_ERROR,
        'WARN_INVLD_OPT_PROP_NETCOM_ENABLED': WARN_INVLD_OPT_PROP_NETCOM_ENABLED,
        'WARN_NOT_CONNECTED': WARN_NOT_CONNECTED,
        'WARN_STLT_NOT_UPDATED': WARN_STLT_NOT_UPDATED,
        'WARN_NO_STLT_CONN_DEFINED': WARN_NO_STLT_CONN_DEFINED,
        'WARN_DEL_UNSET_PROP': WARN_DEL_UNSET_PROP,
        'WARN_RSC_ALREADY_DEPLOYED': WARN_RSC_ALREADY_DEPLOYED,
        'WARN_RSC_ALREADY_HAS_DISK': WARN_RSC_ALREADY_HAS_DISK,
        'WARN_RSC_ALREADY_DISKLESS': WARN_RSC_ALREADY_DISKLESS,
        'WARN_ALL_DISKLESS': WARN_ALL_DISKLESS,
        'WARN_STORAGE_ERROR': WARN_STORAGE_ERROR,
        'WARN_NOT_FOUND_CRYPT_KEY': WARN_NOT_FOUND_CRYPT_KEY,
        'WARN_STORAGE_KIND_ADDED': WARN_STORAGE_KIND_ADDED,
        'WARN_NOT_ENOUGH_NODES_FOR_TIE_BREAKER': WARN_NOT_ENOUGH_NODES_FOR_TIE_BREAKER,
        'WARN_MIXED_PMEM_AND_NON_PMEM': WARN_MIXED_PMEM_AND_NON_PMEM,
        'WARN_UNEFFECTIVE_PROP': WARN_UNEFFECTIVE_PROP,
        'WARN_INVLD_SNAPSHOT_SHIPPING_PREFIX': WARN_INVLD_SNAPSHOT_SHIPPING_PREFIX,
        'WARN_NODE_EVICTED': WARN_NODE_EVICTED,
        'WARN_RSC_DEACTIVATED': WARN_RSC_DEACTIVATED,
        'WARN_BACKUP_DL_ONLY': WARN_BACKUP_DL_ONLY,
        'WARN_NOT_EVACUATING': WARN_NOT_EVACUATING,
        'WARN_INVLD_CONF': WARN_INVLD_CONF,
        'WARN_NOT_FOUND': WARN_NOT_FOUND,
        'INFO_NO_RSC_SPAWNED': INFO_NO_RSC_SPAWNED,
        'INFO_NODE_NAME_MISMATCH': INFO_NODE_NAME_MISMATCH,
        'INFO_PROP_SET': INFO_PROP_SET,
        'INFO_TIE_BREAKER_CREATED': INFO_TIE_BREAKER_CREATED,
        'INFO_TIE_BREAKER_DELETING': INFO_TIE_BREAKER_DELETING,
        'INFO_TIE_BREAKER_TAKEOVER': INFO_TIE_BREAKER_TAKEOVER,
        'INFO_PROP_REMOVED': INFO_PROP_REMOVED,
        'INFO_AUTO_DRBD_PROXY_CREATED': INFO_AUTO_DRBD_PROXY_CREATED,
        'INFO_NOOP': INFO_NOOP,
        'INFO_RSC_ALREADY_EXISTS': INFO_RSC_ALREADY_EXISTS,
        'UNKNOWN_API_CALL': UNKNOWN_API_CALL,
        'API_CALL_AUTH_REQ': API_CALL_AUTH_REQ,
        'API_CALL_PARSE_ERROR': API_CALL_PARSE_ERROR,
        'SUCCESS_SIGN_IN': SUCCESS_SIGN_IN,
        'FAIL_SIGN_IN': FAIL_SIGN_IN,
        'FAIL_SIGN_IN_MISSING_CREDENTIALS': FAIL_SIGN_IN_MISSING_CREDENTIALS
    }
    TYPE_MASKS = [
        0xC000000000000000,  # TYPE
        0x0000000003000000,  # OPERATION
        0x00000000003C0000,  # OBJECT
        0xC00000000000FFFF   # ACTION
    ]

    for num_str in sys.argv[1:]:
        num = long(num_str)
        mask = []
        for type_mask in TYPE_MASKS:
            for key, mask_value in MAP_MASK.items():
                if num & type_mask == mask_value:
                    mask.append(key)
        print(num_str + " = " + " | ". join(mask))



"""
ResourceGroup module
"""

import linstor.errors
from linstor.resource import _Client, Resource


class ResourceGroup(object):
    def __init__(self, name, uri='linstor://localhost', existing_client=None, agent_info=""):
        self._name = name
        self._uri = uri
        self.client = _Client(uri, agent_info=agent_info)
        self._existing_client = existing_client

        self._description = None
        self._redundancy = None
        self._storage_pool_list = None
        self._do_not_place_with = None
        self._do_not_place_with_regex = None
        self._replicas_on_same = None
        self._replicas_on_different = None
        self._diskless_on_remaining = None
        self._layer_list = None
        self._provider_list = None
        self._property_dict = None

        self._nr_volumes = 0
        self._nr_volumes_default = 1

        self._update_or_create()

    def _get_connection(self):
        if self._existing_client:
            return self._existing_client
        return linstor.MultiLinstor(self.client.uri_list, self.client.timeout, self.client.keep_alive,
                                    self.client.agent_info)

    @property
    def name(self):
        """
        Returns the name of the ResourceGroup

        :return: The name of a Resource Group.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        if self._name is not None:
            raise linstor.LinstorReadOnlyAfterSetError('This is a read-only property')
        self._name = name

    @property
    def description(self):
        """
        Returns the description of the ResourceGroup

        :return: The description of a Resource Group.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        self._description = description
        return self._modify_or_create("modify")

    @property
    def redundancy(self):
        """
        Returns the redundancy of the ResourceGroup

        :return: The redundancy of a Resource Group.
        :rtype: int
        """
        return self._redundancy

    @redundancy.setter
    def redundancy(self, redundancy):
        self._redundancy = redundancy
        return self._modify_or_create("modify")

    @property
    def storage_pool(self):
        """
        Returns the storage_pool of the ResourceGroup

        :return: The storage_pool of a Resource Group.
        :rtype: str|list[str]
        """
        return self._storage_pool_list

    @storage_pool.setter
    def storage_pool(self, storage_pool):
        self._storage_pool_list = storage_pool
        return self._modify_or_create("modify")

    @property
    def do_not_place_with(self):
        """
        Returns the do_not_place_with of the ResourceGroup

        :return: The do_not_place_with of a Resource Group.
        :rtype: list[str]
        """
        return self._do_not_place_with

    @do_not_place_with.setter
    def do_not_place_with(self, do_not_place_with):
        self._do_not_place_with = do_not_place_with
        return self._modify_or_create("modify")

    @property
    def do_not_place_with_regex(self):
        """
        Returns the do_not_place_with_regex of the ResourceGroup

        :return: The do_not_place_with_regex of a Resource Group.
        :rtype: str
        """
        return self._do_not_place_with_regex

    @do_not_place_with_regex.setter
    def do_not_place_with_regex(self, do_not_place_with_regex):
        self._do_not_place_with_regex = do_not_place_with_regex
        return self._modify_or_create("modify")

    @property
    def replicas_on_same(self):
        """
        Returns the replicas_on_same of the ResourceGroup

        :return: The replicas_on_same of a Resource Group.
        :rtype: list[str]
        """
        return self._replicas_on_same

    @replicas_on_same.setter
    def replicas_on_same(self, replicas_on_same):
        self._replicas_on_same = replicas_on_same
        return self._modify_or_create("modify")

    @property
    def replicas_on_different(self):
        """
        Returns the replicas_on_different of the ResourceGroup

        :return: The replicas_on_different of a Resource Group.
        :rtype: list[str]
        """
        return self._replicas_on_different

    @replicas_on_different.setter
    def replicas_on_different(self, replicas_on_different):
        self._replicas_on_different = replicas_on_different
        return self._modify_or_create("modify")

    @property
    def diskless_on_remaining(self):
        """
        Returns the diskless_on_remaining of the ResourceGroup

        :return: The diskless_on_remaining of a Resource Group.
        :rtype: bool
        """
        return self._diskless_on_remaining

    @diskless_on_remaining.setter
    def diskless_on_remaining(self, diskless_on_remaining):
        self._diskless_on_remaining = diskless_on_remaining
        return self._modify_or_create("modify")

    def _set_nr_volumes(self, nr_volumes):
        have = self._nr_volumes
        want = nr_volumes
        if have == want:
            return True

        if want < 1:
            raise linstor.LinstorError("A resource group needs at least one volume group")

        with self._get_connection() as lin:
            # inc/dec per interation to keep correct count if we fail in the middle
            if have < want:  # increase
                for v in range(have, want):
                    lin.volume_group_create(self._name, volume_nr=v)
                    self._nr_volumes += 1
            elif have > want:  # decrease
                for v in range(have, want, -1):
                    lin.volume_group_delete(self._name, v - 1)
                    self._nr_volumes -= 1
            # else ==, done

        return True

    @property
    def nr_volumes(self):
        """
        Returns the number of volumes of the ResourceGroup

        :return: The number of volumes of a Resource Group.
        :rtype: int
        """
        return self._nr_volumes

    @nr_volumes.setter
    def nr_volumes(self, nr_volumes):
        return self._set_nr_volumes(nr_volumes)

    @property
    def layer_list(self):
        """
        Returns the layer list of the ResourceGroup

        :return: The layer list of a Resource Group.
        :rtype: list[str]
        """
        return self._layer_list

    @layer_list.setter
    def layer_list(self, layer_list):
        self._layer_list = layer_list
        return self._modify_or_create("modify")

    @property
    def provider_list(self):
        """
        Returns the provider list of the ResourceGroup

        :return: The provider list of a Resource Group.
        :rtype: list[str]
        """
        return self._provider_list

    @provider_list.setter
    def provider_list(self, provider_list):
        self._provider_list = provider_list
        return self._modify_or_create("modify")

    @property
    def property_dict(self):
        """
        Returns the property dict of the ResourceGroup

        :return: The property dict of a Resource Group.
        :rtype: dict[str, str]
        """
        return self._property_dict

    @property_dict.setter
    def property_dict(self, property_dict):
        self._property_dict = property_dict
        return self._modify_or_create("modify")

    def delete(self):
        """
        Deletes the ResourceGroup
        """
        with self._get_connection() as lin:
            lin.resource_group_delete(self._name)
        return True

    def create_resource(self, resource_name, vlm_sizes):
        """
        Create resource with values.

        :param str resource_name: Name of the resource to create.
        :param list[str] vlm_sizes: Volume definitions to spawn
        :return: Resource object of the newly created resource definition
        :rtype: linstor.resource.Resource
        """
        r = Resource.from_resource_group(self._uri, self._name, resource_name, vlm_sizes,
                                         timeout=self.client.timeout, keep_alive=self.client.keep_alive,
                                         existing_client=self._existing_client, agent_info=self.client.agent_info)
        r.client.keep_alive = self.client.keep_alive
        r.client.timeout = self.client.timeout
        r.client.agent_info = self.client.agent_info
        return r

    def query_max_volume_size(self):
        """
        Queries maximum volume size from the given resource group and returns all possible candidates
        """
        with self._get_connection() as lin:
            return lin.resource_group_qmvs(self._name)

    def _modify_or_create(self, what="modify"):
        with self._get_connection() as lin:
            fn = None
            if what == "create":
                fn = lin.resource_group_create
            elif what == "modify":
                fn = lin.resource_group_modify

            fn(self._name,
               description=self._description,
               place_count=self._redundancy,
               storage_pool=self._storage_pool_list,
               do_not_place_with=self._do_not_place_with,
               do_not_place_with_regex=self._do_not_place_with_regex,
               replicas_on_same=self._replicas_on_same,
               replicas_on_different=self._replicas_on_different,
               diskless_on_remaining=self._diskless_on_remaining,
               layer_list=self._layer_list,
               provider_list=self._provider_list,
               property_dict=self._property_dict)

        return True

    def _update(self):
        with self._get_connection() as lin:
            rgs = lin.resource_group_list_raise(filter_by_resource_groups=[self._name]).resource_groups
            rg = rgs[0]
            self._name = rg.name
            self._description = rg.description
            sf = rg.select_filter
            self._redundancy = sf.place_count
            self._storage_pool_list = sf.storage_pool_list
            self._do_not_place_with = sf.not_place_with_rsc
            self._do_not_place_with_regex = sf.not_place_with_rsc_regex
            self._replicas_on_same = sf.replicas_on_same
            self._replicas_on_different = sf.replicas_on_different
            self._diskless_on_remaining = sf.diskless_on_remaining
            self._layer_list = sf.layer_stack
            self._provider_list = sf.provider_list
            self._property_dict = rg.properties

            self._nr_volumes = len(lin.volume_group_list_raise(self._name).volume_groups)

        return True

    def _update_or_create(self):
        with self._get_connection() as lin:
            rgs = lin.resource_group_list_raise(filter_by_resource_groups=[self._name]).resource_groups
            if len(rgs) == 0:  # does not exist yet
                self._modify_or_create("create")
                self._set_nr_volumes(self._nr_volumes_default)

        return self._update()



from .linstorapi import ApiCallResponse, ErrorReport
from .linstorapi import Linstor, MultiLinstor
from .resource import Resource, Volume
from .kv import KV
from .resourcegroup import ResourceGroup
from .config import Config
from .responses import StoragePoolDriver
from .linstorapi import ResourceData
from .size_calc import SizeCalc
from .errors import LinstorError, LinstorTimeoutError, LinstorNetworkError, LinstorApiCallError, LinstorArgumentError
from .errors import LinstorReadOnlyAfterSetError
from . import sharedconsts as consts

try:
    from linstor.consts_githash import GITHASH
except ImportError:
    GITHASH = 'GIT-hash: UNKNOWN'



#!/usr/bin/env python3
import argparse
import sys
import subprocess
import xml.etree.ElementTree as ET


"""
This script translates xml output from drbdsetup options to JSON
"""


_CategoryNamespaces = {
    'new-peer': "DrbdOptions/Net",
    'disk-options': "DrbdOptions/Disk",
    'resource-options': "DrbdOptions/Resource",
    'peer-device-options': "DrbdOptions/PeerDevice"
}

# map a CategoryNamespace to a .res file section
_ResfileSections = {
    'DrbdOptions/Net': 'net',
    'DrbdOptions/Disk': 'disk',
    'DrbdOptions/Resource': 'options',
    'DrbdOptions/PeerDevice': 'disk',
    'DrbdOptions/Handlers': 'handlers'
}

_ObjectCategories = {
    "controller": ['disk-options', 'resource-options', 'new-peer', 'peer-device-options'],
    "resource-definition": ['disk-options', 'resource-options', 'new-peer', 'peer-device-options'],
    "volume-definition": ['disk-options'],  # TODO add volume connection -> 'peer-device-options'
    "rsc-conn": ['peer-device-options', 'new-peer']
}


def get_drbd_setup_xml(from_file):
    if from_file:
        with open(from_file) as f:
            return f.read()

    drbdsetup_cmd = ['/usr/sbin/drbdsetup', 'xml-help']
    opts = ['disk-options', 'peer-device-options', 'resource-options', 'new-peer']
    xml_opts = []
    try:
        xml_opts = [subprocess.check_output(drbdsetup_cmd + [x]) for x in opts]
    except OSError:
        sys.stderr.write("Unable to execute drbdsetup: {cmd}\nUsing local file {f}\n".format(
            cmd=" ".join(drbdsetup_cmd),
            f=from_file)
        )

    return '<root>\n' + "".join(xml_opts) + '</root>'


def create_and_add_handlers_option(properties, option_name):
    if option_name in properties:
        raise RuntimeError("drbd option name already in use: " + option_name)

    cmd_namespace = 'DrbdOptions/Handlers'
    properties[option_name] = {
        'internal': True,
        'key': cmd_namespace + '/' + option_name,
        'drbd_option_name': option_name,
        'drbd_res_file_section': _ResfileSections[cmd_namespace],
        'type': 'string'
    }

    return properties


def add_handlers(objects, properties):
    handlers = [
        "after-resync-target",
        "before-resync-target",
        "before-resync-source",
        "out-of-sync",
        "quorum-lost",
        "fence-peer",
        "unfence-peer",
        "initial-split-brain",
        "local-io-error",
        "pri-lost",
        "pri-lost-after-sb",
        "pri-on-incon-degr",
        "split-brain"
    ]

    for h in handlers:
        create_and_add_handlers_option(properties, h)
        objects["controller"].append(h)
        objects["resource-definition"].append(h)

    return properties


def parse_drbd_setup_xml(xmlout):
    root = ET.fromstring(xmlout)

    objects = {k: [] for k in _ObjectCategories.keys()}
    properties = {}
    for command in root:
        cmd_name = command.attrib['name']
        cmd_namespace = _CategoryNamespaces[cmd_name]

        cmd_properties = {}
        for option in command.findall('option'):
            option_name = option.attrib['name']
            if option_name not in ['set-defaults', '_name']:
                cmd_properties[option_name] = convert_option(cmd_namespace, option_name, option)

        for obj, categories in _ObjectCategories.items():
            if cmd_name in categories:
                objects[obj].extend(cmd_properties.keys())
        properties.update(cmd_properties)

    # add handlers section options
    add_handlers(objects, properties)

    # patch resync-after to be type `string`, the type is still wrong in the drbdsetup.xml
    properties['resync-after']['type'] = "string"
    del properties['resync-after']['max']
    del properties['resync-after']['min']
    del properties['resync-after']['default']
    del properties['resync-after']['unit_prefix']

    return {
        "objects": objects,
        "properties": properties
    }


def convert_option(cmd_namespace, option_name, option):
    option_type = option.attrib['type']

    prop = {
        'internal': True,
        'key': cmd_namespace + '/' + option_name,
        'drbd_option_name': option_name,
        'drbd_res_file_section': _ResfileSections[cmd_namespace]
    }

    if option_type == 'string':
        prop_type = option_type
    elif option_type == 'boolean':
        prop_type = option_type
        prop['default'] = True if option.find('default').text == 'yes' else False
    elif option_type == 'handler':
        prop_type = 'symbol'
        prop['values'] = [h.text for h in option.findall('handler')]
    elif option_type == 'numeric':
        for v in ('unit_prefix', 'unit'):
            val = option.find(v)
            if val is not None:
                prop[v] = val.text
        for v in ['min', 'max', 'default']:
            val = option.find(v)
            if val is not None:
                prop[v] = int(val.text)
        prop_type = 'range' if 'min' in prop.keys() else 'numeric'
    elif option_type == 'numeric-or-symbol':
        prop_type = option_type
        prop['values'] = [h.text for h in option.findall('symbol')]
        prop['min'] = option.find('min').text
        prop['max'] = option.find('max').text
    else:
        raise RuntimeError('Unknown option type ' + option_type)

    prop['type'] = prop_type

    return prop


def gendrbd(output_target):
    xml = get_drbd_setup_xml('drbdsetup.xml')
    props = parse_drbd_setup_xml(xml)
    import json

    with open(output_target, 'wt') as f:
        json.dump(props, f, indent=2, separators=(',', ': '))

    return 0


def main():
    parser = argparse.ArgumentParser(description="generates prepared code containing drbd options")
    parser.add_argument("drbdoptions")

    args = parser.parse_args()

    sys.exit(gendrbd(args.drbdoptions))


if __name__ == '__main__':
    main()



import unittest
from linstor import ApiCallResponse, sharedconsts


class TestResponses(unittest.TestCase):

    def test_ApiCallResponse(self):
        rc_fail_exists = ApiCallResponse.from_json({
            'ret_code': -4611686018407202314,
            'error_report_ids': ['5CFA0EF9-00000-000001'],
            'obj_refs': {'RscDfn': 'test'},
            'details': "Node(s): 'drbd1', Resource: 'test'",
            'message': "A resource 'test' on node 'drbd1' already exists.",
            'cause': 'The Resource already exists'}
        )

        rc_node_not_exits = ApiCallResponse.from_json({
            'message': "Node 'centos1' not found.",
            'cause': "The specified node 'centos1' could not be found in the database",
            'correction': "Create a node with the name 'centos1' first.",
            'ret_code': -4611686018427387604}
        )

        self.assertTrue(rc_fail_exists.is_error())
        self.assertTrue(rc_fail_exists.is_error(sharedconsts.FAIL_EXISTS_RSC))
        self.assertFalse(rc_fail_exists.is_success())

        self.assertFalse(rc_node_not_exits.is_error(sharedconsts.FAIL_EXISTS_RSC))
        self.assertFalse(rc_node_not_exits.is_error(sharedconsts.FAIL_EXISTS_NODE))
        self.assertTrue(rc_node_not_exits.is_error(sharedconsts.FAIL_NOT_FOUND_NODE))



import locale
import re
from .errors import LinstorError
from collections import OrderedDict


class SizeCalc(object):

    """
    Methods for converting decimal and binary sizes of different magnitudes
    """

    _base_2 = 0x0200
    _base_10 = 0x0A00

    UNIT_B = 0 | _base_2
    UNIT_KiB = 10 | _base_2
    UNIT_MiB = 20 | _base_2
    UNIT_GiB = 30 | _base_2
    UNIT_TiB = 40 | _base_2
    UNIT_PiB = 50 | _base_2
    UNIT_EiB = 60 | _base_2
    UNIT_ZiB = 70 | _base_2
    UNIT_YiB = 80 | _base_2

    UNIT_kB = 3 | _base_10
    UNIT_MB = 6 | _base_10
    UNIT_GB = 9 | _base_10
    UNIT_TB = 12 | _base_10
    UNIT_PB = 15 | _base_10
    UNIT_EB = 18 | _base_10
    UNIT_ZB = 21 | _base_10
    UNIT_YB = 24 | _base_10

    """
    Unit keys are lower-case; functions using the lookup table should
    convert the unit name to lower-case to look it up in this table
    """
    UNITS_MAP = OrderedDict([(unit_str.lower(), (unit_str, unit)) for unit_str, unit in [
        ('B', UNIT_B),
        ('K', UNIT_KiB),
        ('kB', UNIT_kB),
        ('KiB', UNIT_KiB),
        ('M', UNIT_MiB),
        ('MB', UNIT_MB),
        ('MiB', UNIT_MiB),
        ('G', UNIT_GiB),
        ('GB', UNIT_GB),
        ('GiB', UNIT_GiB),
        ('T', UNIT_TiB),
        ('TB', UNIT_TB),
        ('TiB', UNIT_TiB),
        ('P', UNIT_PiB),
        ('PB', UNIT_PB),
        ('PiB', UNIT_PiB),
    ]])

    UNITS_LIST_STR = ', '.join([unit_str for unit_str, _ in UNITS_MAP.values()])

    @classmethod
    def unit_to_str(cls, unit):
        for u_key in cls.UNITS_MAP:
            u = cls.UNITS_MAP[u_key]
            if u[1] == unit:
                return u_key
        return ''

    @classmethod
    def parse_unit(cls, value):
        """
        Parses the given value as a computer size + unit.
        If the value doesn't contain a unit indicator, bytes are assumed.

        :param str value: string value to parse
        :return: a tuple of the size value as int and the SizeCalc.UNIT
        :rtype: (int, int)
        """
        m = re.match(r'(\d+)\s*(\D*)', value)

        size = None
        if m:
            size = int(m.group(1))
        else:
            raise LinstorError("Size is not a number.")

        unit_str = m.group(2)
        if unit_str == "":
            unit_str = "B"

        try:
            _, unit = SizeCalc.UNITS_MAP[unit_str.lower()]
        except KeyError:
            raise LinstorError('"%s" is not a valid unit!\n' % unit_str)

        return size, unit

    @classmethod
    def auto_convert(cls, value, unit_to):
        """
        Convertes the given value to another byte unit.
        :param str value: string value that should be converted.
        :param int unit_to: SizeCalc.UNIT enum to convert to
        :return: the converted value
        :rtype: int
        """
        size, unit_from = cls.parse_unit(value)

        return cls.convert(size, unit_from, unit_to)

    @classmethod
    def convert(cls, size, unit_in, unit_out):
        """
        Convert a size value into a different scale unit

        Convert a size value specified in the scale unit of unit_in to
        a size value given in the scale unit of unit_out
        (e.g. convert from decimal megabytes to binary gigabytes, ...)

        :param   size: numeric size value
        :param   unit_in: scale unit selector of the size parameter
        :param   unit_out: scale unit selector of the return value
        :return: size value converted to the scale unit of unit_out
                 truncated to an integer value
        :rtype: int
        """
        fac_in = ((unit_in & 0xffffff00) >> 8) ** (unit_in & 0xff)
        div_out = ((unit_out & 0xffffff00) >> 8) ** (unit_out & 0xff)
        return size * fac_in // div_out

    @classmethod
    def convert_round_up(cls, size, unit_in, unit_out):
        """
        Convert a size value into a different scale unit and round up

        Convert a size value specified in the scale unit of unit_in to
        a size value given in the scale unit of unit_out
        (e.g. convert from decimal megabytes to binary gigabytes, ...).
        The result is rounded up so that the returned value always specifies
        a size that is large enough to contain the size supplied to this
        function.
        (e.g., for 100 decimal Megabytes (MB), which equals 100 million bytes,
         returns 97,657 binary kilobytes (kiB), which equals 100 million
         plus 768 bytes and therefore is large enough to contain 100 megabytes)

        :param   size: numeric size value
        :param   unit_in: scale unit selector of the size parameter
        :param   unit_out: scale unit selector of the return value
        :return: size value converted to the scale unit of unit_out
        :rtype: int
        """
        fac_in = ((unit_in & 0xffffff00) >> 8) ** (unit_in & 0xff)
        div_out = ((unit_out & 0xffffff00) >> 8) ** (unit_out & 0xff)
        byte_sz = size * fac_in
        if byte_sz % div_out != 0:
            result = (byte_sz / div_out) + 1
        else:
            result = byte_sz / div_out
        return int(result)

    @classmethod
    def approximate_size_string(cls, size_kib):
        """
        Produce human readable size information as a string

        :param int size_kib: size in kibi byte
        :return: human readable size string
        :rtype: str
        """
        units = [
            "KiB",
            "MiB",
            "GiB",
            "TiB",
            "PiB"
        ]
        max_index = len(units)

        index = 0
        counter = 1
        magnitude = 1 << 10
        while counter < max_index:
            if size_kib >= magnitude:
                index = counter
            else:
                break
            magnitude = magnitude << 10
            counter += 1
        magnitude = magnitude >> 10

        size_str = None
        if size_kib % magnitude != 0:
            size_unit = float(size_kib) / magnitude
            size_loc = locale.format_string('%3.2f', size_unit, grouping=True)
            size_str = "%s %s" % (size_loc, units[index])
        else:
            size_unit = size_kib / magnitude
            size_str = "%d %s" % (size_unit, units[index])

        return size_str



class LinstorError(Exception):
    """
    Linstor basic error class with a message
    """
    def __init__(self, msg, all_errors=None):
        self._msg = msg
        if all_errors is None:
            all_errors = []
        self._errors = all_errors

    def all_errors(self):
        return self._errors

    @property
    def message(self):
        return self._msg

    def __str__(self):
        return "Error: {msg}".format(msg=self._msg)

    def __repr__(self):
        return "LinstorError('{msg}')".format(msg=self._msg)


class LinstorNetworkError(LinstorError):
    """
    Linstor Error indicating an network/connection error.
    """
    def __init__(self, msg, more_errors=None):
        super(LinstorNetworkError, self).__init__(msg, more_errors)


class LinstorTimeoutError(LinstorError):
    """
    Linstor network timeout error
    """
    def __init__(self, msg, more_errors=None):
        super(LinstorTimeoutError, self).__init__(msg, more_errors)


class LinstorApiCallError(LinstorError):
    """
    Linstor error from an apicall response.
    """
    def __init__(self, apicallresponse, all_errors=None):
        """

        :param ApiCallResponse apicallresponse:
        :param list[ApiCallResponse] all_errors:
        """
        super(LinstorApiCallError, self).__init__(str(apicallresponse), all_errors if all_errors else [apicallresponse])

        self._main_error = apicallresponse

    @property
    def main_error(self):
        return self._main_error


class LinstorArgumentError(LinstorError):
    """
    Linstor error if an argument for a function call is invalid.
    """
    def __init__(self, msg, more_errors=None):
        super(LinstorArgumentError, self).__init__(msg, more_errors)


class LinstorReadOnlyAfterSetError(LinstorError):
    """
    Linstor error raised if a property that is only allowed to be set once, is re-set.
    """
    def __init__(self, msg=None, more_errors=None):
        if msg is None:
            msg = 'After this property got set it is read-only'
        super(LinstorReadOnlyAfterSetError, self).__init__(msg, more_errors)



import unittest

if __name__ == '__main__':
    import xmlrunner
    loader = unittest.TestLoader()
    suite = loader.discover(start_dir='linstor_tests')
    runner = xmlrunner.XMLTestRunner(output='test-reports')
    runner.run(suite)



#!/usr/bin/env python3
from __future__ import print_function

import argparse
import json
import datetime
import os
import copy
import pprint
import re
import sys


if sys.version_info > (3, 0):
    PYTHON2 = False
    unicode = str

basename = os.path.basename(__file__)
PROPERTY_JSON_PATH = os.path.join(sys.argv[2] if len(sys.argv) > 1 else __file__)
COPY_YEAR = datetime.datetime.fromtimestamp(os.path.getmtime(PROPERTY_JSON_PATH)).year
hdr = 'This file was autogenerated by %s' % basename
license = """
LINSTOR - management of distributed storage/DRBD9 resources
Copyright (C) 2017 - %s  LINBIT HA-Solutions GmbH
Author: %s

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.""" % (
    COPY_YEAR, ', '.join(['Rene Peinthor', 'Gabor Hernadi']))


class MyPyKey(object):
    def __init__(self, keypath):
        self._keys = keypath

    def __str__(self):
        if isinstance(self._keys, list):
            return " + '/' + ".join(['consts.' + x for x in self._keys])
        return '"{k}"'.format(k=self._keys)

    def __repr__(self):
        return str(self)


def merge_props(prop_a, prop_b):
    prps = copy.deepcopy(prop_a)
    prps['properties'].update(prop_b['properties'])

    for obj in prop_b['objects']:
        for e in prop_b['objects'].get(obj, []):
            if obj not in prps['objects']:
                prps['objects'][obj] = []
            prps['objects'][obj].append(e)

    return prps


def check_duplicate_keys(propdata):
    with open("consts.json") as constfile:
        constjson = json.load(constfile)
        consts = {e['name']: e['value'] for e in constjson if e.get('type') == 'string'}

    prop_map = {}
    for prop_name in propdata['properties']:
        key = propdata['properties'][prop_name]['key']
        if isinstance(key, list):
            key = "/".join([consts[x] for x in key])
        if prop_name in prop_map:
            print("ERROR: DUPLICATE property name '{k}' for prop '{p}'".format(k=key, p=prop_name), file=sys.stderr)
            sys.exit(9)
        prop_map[prop_name] = key

    for obj in propdata['objects']:
        key_map = {}
        for prop_name in propdata['objects'][obj]:
            key = prop_map[prop_name]
            if key in key_map:
                print("ERROR: DUPLICATE property key '{k}' for prop '{p}' in obj '{o}'".format(
                    k=key, p=prop_name, o=obj), file=sys.stderr)
                sys.exit(9)
            key_map[prop_map[prop_name]] = prop_name


def check_info_prop(propdata):
    """
    Checks if all non internal properties do have an info string
    :param propdata: dict with all properties
    :return:
    """
    for pn in propdata['properties']:
        prop = propdata['properties'][pn]
        if not prop.get('internal', False) and 'info' not in prop:
            print("ERROR: Public property '{p}' does not have an info.".format(p=pn), file=sys.stderr)
            sys.exit(8)


def lang_python(data):
    print('"""')
    print('{h}\n{lic}'.format(h=hdr, lic=license))
    print('"""')
    print('import linstor.sharedconsts as consts')

    properties = data['properties']
    objects = data['objects']
    p = {}

    # prefix keys with consts.
    for prop in properties:
        key = properties[prop]['key']
        properties[prop]['key'] = MyPyKey(key)

    print('properties = \\')
    for objkey in objects:
        obj_props = objects[objkey]
        resolved = []
        for prop in obj_props:
            resolved.append(properties[prop])

        p[objkey] = resolved

    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(p)
    return True


def lang_java(data):
    properties = data['properties']
    objects = data['objects']

    license_hdr = ''
    for line in license.split('\n'):
        license_hdr += (' * ' + line).rstrip() + '\n'

    print('/*\n * %s\n%s */\n' % (hdr, license_hdr))

    _print(0, 'package com.linbit.linstor.api.prop;')
    _print(0, '')
    _print(0, 'import java.util.ArrayList;')
    _print(0, 'import java.util.Arrays;')
    _print(0, 'import java.util.List;')
    _print(0, 'import java.util.Map;')
    _print(0, 'import java.util.TreeMap;')
    _print(0, '')
    _print(0, 'public final class GeneratedPropertyRules\n{')
    _print(1,     'public static List<Property> getWhitelistedProperties()')
    _print(1,     '{')
    _print(2,         'List<Property> propertyList = new ArrayList<>();')
    for key in sorted(properties.keys()):
        prop = properties[key]
        _print(2,     'propertyList.add(')
        _print(3,         'new PropertyBuilder()')
        _print(4,             '.name("%s")' % key)
        for propKey in sorted(prop.keys()):
            if _relevant_for_java(propKey):
                propVal = prop[propKey]
                if isinstance(propVal, unicode):
                    propVal = propVal.replace('"', '\\"')
                if propKey == "key":
                    if isinstance(propVal, list):
                        _print(4, '.keyRef("%s")' % ('", "'.join(propVal)))
                    else:
                        _print(4, '.keyStr("%s")' % (propVal))
                elif propKey == "values":
                    _print(4,     '.values(')
                    _print(5,         '"%s"' % ('",\n' + _indent(5) + '"').join(sorted(propVal)))
                    _print(4,     ')')
                elif propKey == "default":
                    _print(4, '.dflt("%s")' % (propVal))
                else:
                    _print(4,     '.%s("%s")' % (propKey, propVal))
        _print(4,             '.build()')
        _print(2,     ');')
    _print(2,         'return propertyList;')
    _print(1,     '}\n')
    _print(1,     'public static Map<LinStorObject, List<String>> getWhitelistedRules()')
    _print(1,     '{')
    _print(2,         'Map<LinStorObject, List<String>> rules = new TreeMap<>();')
    for obj_name in sorted(objects.keys()):
        rule_names = sorted(objects[obj_name])
        if not rule_names:
            _print(2, 'rules.put(LinStorObject.%s, Collections.emptyList());' % _as_java_enum_name(obj_name))
        else:
            _print(2, 'rules.put(LinStorObject.%s,' % _as_java_enum_name(obj_name))
            _print(3,     'Arrays.asList(')
            _print(4,         '"%s"' % ('",\n' + _indent(4) + '"').join(rule_names))
            _print(3,     ')')
            _print(2, ');')
    _print(2,         'return rules;')
    _print(1,     '}')
    _print(0, '}\n')
    return True


def _print(indent_level, line):
    indent = _indent(indent_level)
    print('%s%s' % (indent, line.strip()))


def _indent(indent_level):
    return ' ' * 4 * indent_level


def _relevant_for_java(propKey):
    return propKey not in ['drbd_option_name', 'unit_prefix', 'drbd_res_file_section']


def _as_java_rule_name(name):
    return re.sub(r'_(.)', lambda pat: pat.group(1).upper(), name) + 'Rule'


def _as_java_enum_name(name):
    return re.sub(r'-', '_', name).upper()


def main():
    parser = argparse.ArgumentParser(prog="genproperties.py")
    parser.add_argument('language', choices=['python', 'java'])
    parser.add_argument('propfile', nargs='+')

    args = parser.parse_args()

    propdata = {
        "properties": {},
        "objects": {}
    }

    for propfile_name in args.propfile:
        with open(propfile_name) as propfile:
            propdata = merge_props(propdata, json.load(propfile))

    # check duplicate keys
    check_duplicate_keys(propdata)

    # check info prop
    check_info_prop(propdata)

    if args.language == 'python':
        lang_python(propdata)
    elif args.language == 'java':
        lang_java(propdata)


if __name__ == '__main__':
    main()



"""
Resource module
"""

import socket
import sys
import random
from functools import wraps

import linstor.linstorapi
from linstor.sharedconsts import FLAG_DISKLESS
from linstor.responses import ResourceDefinitionResponse, ResourceResponse
from linstor.linstorapi import Linstor

PYTHON2 = True
if sys.version_info > (3, 0):
    PYTHON2 = False
    unicode = str


class _Utils(object):
    @classmethod
    def to_unicode(cls, t):
        if isinstance(t, str):
            if PYTHON2:
                return unicode(t, 'UTF-8')
            else:
                return t
        elif isinstance(t, unicode):
            return t
        else:
            if PYTHON2:
                return unicode(t)
            else:
                return str(t)


class _Client(object):
    def __init__(self, uris, timeout=300, keep_alive=False, agent_info=""):
        # external properties
        self._uri_list = linstor.MultiLinstor.controller_uri_list(uris)  # type: list[str]
        self.timeout = timeout
        self.keep_alive = keep_alive
        self.agent_info = agent_info

    @property
    def uri_list(self):
        return self._uri_list

    @uri_list.setter
    def uri_list(self, uri_list):
        raise linstor.LinstorReadOnlyAfterSetError()


class _Placement(object):
    def __init__(self, redundancy=None):
        self.redundancy = redundancy
        self.storage_pool = None
        self.diskless_storage_pool = None
        self.diskless_on_remaining = False


class Volume(object):
    """
    Volume class represents a DRBD Volume.

    This object contains important properties of a Volume, including
    e.g: ``size``, and ``device_path``

    :param size: String parsable by linstor.SizeCalc or size in bytes.
    """
    def __init__(self, size):
        # external properties
        self._size = self._size_to_bytes(size)
        self._minor = None
        self._device_path = ''
        self._storage_pool_name = ''

        # internal
        self._volume_id = None
        self._rsc_name = None
        self._client_ref = None  # type: Optional[linstor.Linstor]
        self._assignments = []

    def __repr__(self):
        return "Volume({n}, {nr}, {s}kib, {m})".format(
            n=self._rsc_name, nr=self._volume_id, s=self._size, m=self._minor
        )

    @property
    def storage_pool_name(self):
        return self._storage_pool_name

    @property
    def device_path(self):
        """
        Returns the device path of a Volume (e.g., /dev/drbd1000).

        :return: The device path of a Volume.
        :rtype: str
        """
        return self._device_path

    @device_path.setter
    def device_path(self, device_path):
        raise linstor.LinstorReadOnlyAfterSetError('This is a read-only property')

    @property
    def minor(self):
        """
        Returns the minor number of a Volume (e.g., 1000).

        :return: The minor number of a Volume.
        :rtype: int
        """
        return self._minor

    @minor.setter
    def minor(self, minor):
        if self._rsc_name is not None:
            raise linstor.LinstorReadOnlyAfterSetError()
        self._minor = minor

    @property
    def size(self):
        """
        Returns the size of a Volume (e.g., 1000000).

        Setting tye size of a volume that is deployed triggers a resize operation.

        :return: The size of a Volume in bytes.
        :rtype: int
        """
        return self._size

    @classmethod
    def _size_to_bytes(cls, size):
        if isinstance(size, str):
            return linstor.SizeCalc.auto_convert(size, linstor.SizeCalc.UNIT_B)
        return size

    @size.setter
    def size(self, size):  # this has to be an int, otherwise python complains
        size = self._size_to_bytes(size)
        if self._size is not None \
           and self._rsc_name is not None \
           and self._volume_id is not None:
            r, v = self._rsc_name, self._volume_id
            if self._size > size:
                raise ValueError('shrinking Resource/Volume {}/{} from {} to {} is not allowed'
                                 .format(r, v, self._size, size))

            size_kib = linstor.SizeCalc.convert_round_up(size, linstor.SizeCalc.UNIT_B,
                                                         linstor.SizeCalc.UNIT_KiB)
            with self._client_ref as lin:
                rs = lin.volume_dfn_modify(r, v, size=size_kib)
                if not linstor.Linstor.all_api_responses_no_error(rs):
                    raise linstor.LinstorError('Could not resize Resource/Volume {}/{}: {}'
                                               .format(r, v, Linstor.filter_api_call_response_errors(rs)[0]))

        # if we are here everything is fine
        self._size = size

    # called from VolumeDict
    def _delete(self):
        if self._rsc_name is None:  # this volume was created, but never deployed, no linstor action.
            return
        with self._client_ref as lin:
            r, v = self._rsc_name, self._volume_id
            rs = lin.volume_dfn_delete(r, v)
            if not linstor.Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError('Could not delete Resource/Volume {}/{}: {}'.format(
                    r, v, Linstor.filter_api_call_response_errors(rs)[0]))


class _VolumeDict(dict):
    def __init__(self):
        super(_VolumeDict, self).__init__()

    def __setitem__(self, k, v):
        if not isinstance(v, Volume):
            raise ValueError('{} is not of an instance of Volume'.format(v))
        v._volume_id = k
        super(_VolumeDict, self).__setitem__(k, v)

    def __delitem__(self, k):
        self[k]._delete()
        super(_VolumeDict, self).__delitem__(k)


class Resource(object):
    """
    Resource class represents a DRBD Resource.

    This object allows managing existing DRBD Resources as well as creating new ones.

    :param str name: The name of the DRBD resource.
    :param str uri: A list of controller addresses.
     e.g: ``linstor://localhost,10.0.0.2``, ``linstor+ssl://localhost,linstor://192.168.0.1``
    :param linstor.Linstor existing_client: Instead of creating  a new client based on the controller addresses,
     use this pre-configured client object.
    """
    def _update_volumes(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            ret = None
            with self._get_connection() as lin:
                self._lin = lin
                self._maybe_create_rd_and_vd()
                ret = f(self, *args, **kwargs)
                self.__update_volumes()
            self._lin = None
            return ret
        return wrapper

    def __init__(self, name, uri='linstor://localhost', existing_client=None, resource_group=None):
        # external properties
        self._name = name  # the user facing name, what linstor calls the "external name"
        self._port = None
        self._resource_group_name = resource_group  # type: Optional[str]
        self.client = _Client(uri)
        self.placement = _Placement()
        self.volumes = _VolumeDict()  # type: dict[int, Volume]
        self.defined = False

        # THINK(rck): maybe a dict, KISS for now
        self._allow_two_primaries = False

        # internal
        self._assignments = {}
        self._linstor_name = None
        self._existing_client = existing_client

        with self._get_connection() as lin:
            self._lin = lin
            self.__update_volumes()

    def __str__(self):
        return '{e}({i})'.format(e=self._name, i=self._linstor_name)

    def __repr__(self):
        return "Resource({n}, {h})".format(n=self, h=self.client.uri_list)

    @classmethod
    def from_resource_group(cls, uri, resource_group_name, resource_name, vlm_sizes,
                            timeout=300, keep_alive=False, definitions_only=False, existing_client=None, agent_info=""):
        """
        Spawns a new resource definition from the given resource group.

        :param str uri: A list of controller addresses.
         e.g: ``linstor://localhost,10.0.0.2``, ``linstor+ssl://localhost,linstor://192.168.0.1``
        :param str resource_group_name: Name of the resource group
        :param str resource_name: Name of the new resource definition
        :param list[str] vlm_sizes: String list of volume sizes e.g. ['128Mib', '1G']
        :param int timeout: client library timeout
        :param bool keep_alive: keep client connection alive
        :param bool definitions_only: only spawn definitions
        :param linstor.Linstor existing_client: Client to associate with the resource
        :param str agent_info: Info string added to user-agent
        :return: linstor.resource.Resource object of the newly created resource definition
        :rtype: linstor.resource.Resource
        """
        if existing_client:
            client = existing_client
        else:
            c = _Client(uri)
            client = linstor.MultiLinstor(c.uri_list, timeout, keep_alive, agent_info)

        with client as lin:
            result = lin.resource_group_spawn(
                resource_group_name,
                resource_name,
                vlm_sizes,
                definitions_only=definitions_only
            )
            if not linstor.Linstor.all_api_responses_no_error(result):
                raise linstor.LinstorError(
                    'Could not spawn resource "{}" from resource group "{}": {}'.format(
                        resource_name,
                        resource_group_name,
                        Linstor.filter_api_call_response_errors(result)[0].message
                    )
                )

            resource = Resource(resource_name, uri=uri, existing_client=existing_client)
            if definitions_only:
                rsc_grp = lin.resource_group_list_raise(filter_by_resource_groups=[resource_group_name])
                stor_pool = None
                if rsc_grp.resource_groups[0].select_filter.storage_pool_list:
                    stor_pool = random.choice(rsc_grp.resource_groups[0].select_filter.storage_pool_list)
                resource.placement.storage_pool = stor_pool
            return resource
        return None

    def _get_connection(self):
        if self._existing_client:
            return self._existing_client
        return linstor.MultiLinstor(self.client.uri_list, self.client.timeout, self.client.keep_alive,
                                    self.client.agent_info)

    def _set_properties(self):
        dp = 'yes' if self._allow_two_primaries else 'no'
        props = {'DrbdOptions/Net/allow-two-primaries': dp}
        rs = self._lin.resource_dfn_modify(self._linstor_name, props, delete_props=None)
        if not linstor.Linstor.all_api_responses_no_error(rs):
            raise linstor.LinstorError('Could not set DRBD properties for resource {}: {}'
                                       .format(self, Linstor.filter_api_call_response_errors(rs)[0]))

    def _maybe_create_rd_and_vd(self):
        # resource definition
        if not self.defined:
            rs = self._lin.resource_dfn_create(
                "", self._port, external_name=self._name, resource_group=self._resource_group_name)
            if not linstor.Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError('Could not create resource definition {}: {}'
                                           .format(self, Linstor.filter_api_call_response_errors(rs)[0]))
            ors = rs[0].object_refs
            try:
                self._linstor_name = ors['RscDfn']
            except KeyError:
                raise linstor.LinstorError('Could not get RscDfn for resource definition {}'
                                           .format(self))

            self.defined = True
            self._set_properties()

        # volume definitions
        self._create_volume_definitions()

    def _create_volume_definitions(self):
        for k, v in self.volumes.items():
            if v._rsc_name is None:
                size_kib = linstor.SizeCalc.convert_round_up(v.size, linstor.SizeCalc.UNIT_B,
                                                             linstor.SizeCalc.UNIT_KiB)
                rs = self._lin.volume_dfn_create(self._linstor_name, size_kib, k, v._minor,
                                                 encrypt=False, storage_pool=self.placement.storage_pool)
                if not linstor.Linstor.all_api_responses_no_error(rs):
                    raise linstor.LinstorError('Could not create volume definition {n}/{k}: {e}'
                                               .format(n=self, k=k, e=Linstor.filter_api_call_response_errors(rs)[0]))
                self.volumes[k]._rsc_name = self._linstor_name

    def __update_volumes(self):
        # create fresh volume definitions
        self._create_volume_definitions()

        # update internal state
        rsc_dfn_list_replies = self._lin.resource_dfn_list()
        if not rsc_dfn_list_replies or not rsc_dfn_list_replies[0]:
            return True

        rsc_dfn_list_reply = rsc_dfn_list_replies[0]  # type: ResourceDefinitionResponse
        for rsc_dfn in rsc_dfn_list_reply.resource_definitions:
            # WORKAROUND: linstor-server < 0.9.9 did not set the external_name, so for compat
            # and as these only used non external names, fall back to the name
            to_cmp = rsc_dfn.external_name if rsc_dfn.external_name != "" else rsc_dfn.name
            if _Utils.to_unicode(to_cmp) == _Utils.to_unicode(self._name):
                self._linstor_name = rsc_dfn.name
                self._resource_group_name = rsc_dfn.resource_group_name
                self.defined = True
                for vlm_dfn in rsc_dfn.volume_definitions:
                    vlm_nr = vlm_dfn.number
                    if not self.volumes.get(vlm_nr):
                        self.volumes[vlm_nr] = Volume(None)
                    self.volumes[vlm_nr]._volume_id = vlm_nr
                    self.volumes[vlm_nr]._rsc_name = self._linstor_name
                    self.volumes[vlm_nr]._client_ref = self._get_connection()
                    size_b = linstor.SizeCalc.convert_round_up(vlm_dfn.size, linstor.SizeCalc.UNIT_KiB,
                                                               linstor.SizeCalc.UNIT_B)
                    self.volumes[vlm_nr]._size = size_b
                    if vlm_dfn.drbd_data is not None:
                        self.volumes[vlm_nr]._minor = vlm_dfn.drbd_data.minor
                for key, value in rsc_dfn.properties.items():
                    if key == 'DrbdOptions/Net/allow-two-primaries':
                        self._allow_two_primaries = True if value == 'yes' else False

        if self._linstor_name is None:
            return True

        rsc_list_replies = self._lin.resource_list(filter_by_nodes=None,
                                                   filter_by_resources=[self._linstor_name])
        if not rsc_list_replies or not rsc_list_replies[0]:
            return True

        self._assignments = {}
        rsc_list_reply = rsc_list_replies[0]  # type: ResourceResponse
        for rsc in rsc_list_reply.resources:
            is_diskless = (FLAG_DISKLESS in rsc.flags)
            node_name = rsc.node_name
            self._assignments[node_name] = is_diskless
            for vlm in rsc.volumes:
                vlm_nr = vlm.number
                if vlm.device_path:
                    self.volumes[vlm_nr]._device_path = vlm.device_path
                if vlm.storage_pool_name and not is_diskless:
                    self.volumes[vlm_nr]._storage_pool_name = vlm.storage_pool_name
                if vlm.drbd_data is not None:
                    self.volumes[vlm_nr]._minor = vlm.drbd_data.drbd_volume_definition.minor

        return True

    @property
    def allow_two_primaries(self):
        """
        Returns the value of the DRBD net-option 'allow-two-primaries'.

        :return: The value of the DRBD net-option 'allow-two-primaries'. Raises LinstorError in case of error.
        :rtype: bool
        """
        return self._allow_two_primaries

    @allow_two_primaries.setter
    def allow_two_primaries(self, value):
        if self._allow_two_primaries == value:
            return

        self._allow_two_primaries = value
        if self.defined:
            with self._get_connection() as lin:
                self._lin = lin
                self._set_properties()

    @property
    def name(self):
        """
        Returns the external/user facing name of the Resource.

        :return: The external/user facing name of the Resource.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        if self.defined:
            raise linstor.LinstorReadOnlyAfterSetError()
        self._name = name

    @property
    def linstor_name(self):
        """
        Returns the internal/linstor/DRBD name of the Resource.

        :return: The internal/linstor/DRBD name of the Resource.
        :rtype: str
        """
        return self._linstor_name

    @linstor_name.setter
    def linstor_name(self, name):
        raise linstor.LinstorReadOnlyAfterSetError()

    @property
    def port(self):
        """
        Returns the port of the Resource.

        :return: The port of the Resource.
        :rtype: str
        """
        return self._port

    @port.setter
    def port(self, port_nr):
        if self.defined:
            raise linstor.LinstorReadOnlyAfterSetError()
        self._port = port_nr

    @property
    def resource_group_name(self):
        return self._resource_group_name

    @_update_volumes
    def autoplace(self):
        """
        Automatically place the Resource according to values set in the placement policy.

        Example:
            To autoplace a Resource 'foo' 3 times redundant on the storage pool 'drbdpool' one would::

                $ foo.placement.redundancy = 3
                $ foo.placement.storage_pool = 'drbdpool'
                $ foo.autoplace()

        :return: True if success, else raises LinstorError
        """
        rs = self._lin.resource_auto_place(
            self._linstor_name,
            self.placement.redundancy,
            self.placement.storage_pool,
            do_not_place_with=None,
            do_not_place_with_regex=None,
            replicas_on_same=None,
            replicas_on_different=None,
            diskless_on_remaining=self.placement.diskless_on_remaining)

        if not Linstor.all_api_responses_no_error(rs):
            raise linstor.LinstorError('Could not autoplace resource {}: {}'
                                       .format(self, Linstor.filter_api_call_response_errors(rs)[0]))
        return True

    @_update_volumes
    def activate(self, node_name, diskful=False):
        """
        Makes a resource available at a given host.

        If the host already contains a diskful assignment, this is a NOOP. Otherwise a diskless assignment is
        created.

        :param str node_name: Name of the node
        :param bool diskful: Make the resource diskful on the node
        :return: True if success, else raises LinstorError
        """
        rsc_available_replies = self._lin.resource_make_available(node_name, self._linstor_name, diskful)

        if not Linstor.all_api_responses_no_error(rsc_available_replies):
            raise linstor.LinstorError('Could not make resource {} available on node {}: {}'
                                       .format(self, node_name, ";".join([str(x) for x in rsc_available_replies])))
        return True

    # no decorator, calles delete
    def deactivate(self, node_name):
        """
        Deactivates a resource on a host if possible.

        If the assignment is diskless, delete this assignment. If it is diskful and therefore part of the
        given redundany, this is a NOOP (i.e., the redundancy is not decreased).

        :param str node_name: Name of the node
        :return: True if success, else raises LinstorError
        """
        if self.is_diskless(node_name):
            return self.delete(node_name)
        return True

    @_update_volumes
    def _create_or_toggle(self, node_name, diskless):
        """

        :param str node_name:
        :param bool diskless:
        :return: True if it was successful
        :rtype: bool
        """
        is_assigned = self.is_assigned(node_name)
        is_diskless = self.is_diskless(node_name)
        sp = self.placement.diskless_storage_pool
        if is_diskless or (not is_assigned and not diskless):
            sp = self.placement.storage_pool

        if not is_assigned:
            rs = self._lin.resource_create([
                linstor.ResourceData(
                    node_name,
                    self._linstor_name,
                    diskless=diskless,
                    storage_pool=sp
                )
            ])
            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError(
                    'Could not create resource {} on node {} as diskless={}: {}'
                    .format(self, node_name, diskless, Linstor.filter_api_call_response_errors(rs)[0]))
        elif is_diskless != diskless:
            rs = self._lin.resource_toggle_disk(node_name, self._linstor_name,
                                                diskless=diskless, storage_pool=sp)
            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError(
                    'Could not toggle disk for resource {} on node {} to diskless={}: {}'
                    .format(self, node_name, diskless, Linstor.filter_api_call_response_errors(rs)[0]))
        return True

    def snapshot_create(self, name):
        """
        Creates a new snapshot for the resource.

        :param str name: Name of the snapshot
        :return: True if success, else raises LinstorError
        """
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        with self._get_connection() as lin:
            rs = lin.snapshot_create(node_names=[], rsc_name=self._linstor_name, snapshot_name=name, async_msg=False)
            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError('Could not create snapshot {}: {}'
                                           .format(name, Linstor.filter_api_call_response_errors(rs)[0].message))
        return True

    def snapshot_delete(self, name):
        """
        Deletes a given snapshot of this resource.

        :param str name: Name of the snapshot
        :return: True if success, else raises LinstorError
        """
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        with self._get_connection() as lin:
            rs = lin.snapshot_delete(rsc_name=self._linstor_name, snapshot_name=name)
            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError('Could not delete snapshot {}: {}'.format(
                    name,
                    Linstor.filter_api_call_response_errors(rs)[0].message))
        return True

    def snapshot_rollback(self, name):
        """
        Rolls resource data back to snapshot state. The resource must not be in use.
        The snapshot will not be removed and can be used for subsequent rollbacks.
        Only the most recent snapshot may be used; to roll back to an earlier
        snapshot, the intermediate snapshots must first be deleted.

        :param str name: Name of the snapshot
        :return: True if success, else raises LinstorError
        """
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        with self._get_connection() as lin:
            rs = lin.snapshot_rollback(rsc_name=self._linstor_name, snapshot_name=name)
            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError('Could not rollback to snapshot {}: {}'.format(
                    name,
                    Linstor.filter_api_call_response_errors(rs)[0].message)
                )
        return True

    def restore_from_snapshot(self, snapshot_name, resource_name_to):
        """
        Restores a new resource from a snapshot.

        :param snapshot_name: Snapshot name to use for restoration.
        :param resource_name_to: Name of the new resource.
        :return: A new resource object restored from the snapshot.
        :rtype: linstor.resource.Resource
        """
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        with self._get_connection() as lin:
            rs = lin.resource_dfn_create(resource_name_to, resource_group=self.resource_group_name)
            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError(
                    "Could not resource definition '{r}' for snapshot restore: {err}"
                    .format(r=resource_name_to, err=Linstor.filter_api_call_response_errors(rs)[0].message))

            rs = lin.snapshot_volume_definition_restore(
                from_resource=self._linstor_name,
                from_snapshot=snapshot_name,
                to_resource=resource_name_to
            )

            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError(
                    "Could not restore volume definition '{rd}' from snapshot {sn} to resource definition '{tr}': {err}"
                    .format(
                        rd=self._linstor_name,
                        sn=snapshot_name,
                        tr=resource_name_to,
                        err=Linstor.filter_api_call_response_errors(rs)[0].message
                    )
                )

            rs = lin.snapshot_resource_restore(
                node_names=[],  # to all
                from_resource=self._linstor_name,
                from_snapshot=snapshot_name,
                to_resource=resource_name_to
            )

            if not Linstor.all_api_responses_no_error(rs):
                raise linstor.LinstorError(
                    "Could not restore resource '{rd}' from snapshot {sn} to resource definition '{tr}': {err}"
                    .format(
                        rd=self.name,
                        sn=snapshot_name,
                        tr=resource_name_to,
                        err=Linstor.filter_api_call_response_errors(rs)[0].message
                    )
                )

        return Resource(resource_name_to, ",".join(self.client.uri_list), existing_client=self._existing_client)

    def clone(self, clone_name, clone_external_name=None, timeout=None, use_zfs_clone=None):
        """
        Starts a clone operation on the current resource to the given clone_name.

        :param str clone_name: Clone name
        :param Optional[str] clone_external_name: Clone external name
        :param Optional[int] timeout: How long to wait for the clone to finish (secs)
        :param Optional[bool] use_zfs_clone: use fast zfs clone with snapshot (dependent clone)
        :return: A new resource object with the cloned resource
        :rtype: linstor.resource.Resource
        :raise LinstorError: On errors
        """
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        with self._get_connection() as lin:
            c_started = lin.resource_dfn_clone(self._linstor_name, clone_name, clone_external_name, use_zfs_clone)

            if not Linstor.all_api_responses_no_error(c_started.messages):
                raise linstor.LinstorError(
                    "Could not clone resource definition '{r}': {err}"
                    .format(r=self._linstor_name,
                            err=Linstor.filter_api_call_response_errors(c_started.messages)[0].message))

            lin.resource_dfn_clone_wait_complete(c_started.source_name, c_started.clone_name, timeout=timeout)

            return Resource(c_started.clone_name, ",".join(self.client.uri_list), existing_client=self._existing_client)

    def diskless(self, node_name):
        """
        Assign a resource diskless on a given node.

        If the assignment does not exist, create it diskless. If the assignment is already diskless, this is a
        NOOP. If it exists diskful, convert it to diskless.

        :param str node_name: Name of the node
        :return: True if success, else raises LinstorError
        """
        return self._create_or_toggle(node_name, True)

    def diskful(self, node_name):
        """
        Assign a resource diskful on a given node.

        If the assignment does not exist, create it diskful. If the assignment is already diskful, this is a
        NOOP. If it exists diskless, convert it to diskful.

        :param str node_name: Name of the node
        :return: True if success, else raises LinstorError
        """
        return self._create_or_toggle(node_name, False)

    def is_diskless(self, node_name):
        """
        Returns True if the resource is assigned diskless on the given host.

        :param str node_name: Name of the node
        :return: True if assigned diskless on given host.
        :rtype: bool
        """
        return self._assignments.get(node_name, False)

    def is_diskful(self, node_name):
        """
        Returns True if the resource is assigned diskful on the given host.

        :param str node_name: Name of the node
        :return: True if assigned diskful on given host.
        :rtype: bool
        """
        return not self._assignments.get(node_name, True)

    def is_assigned(self, node_name):
        """
        Returns True if the resource is assigned diskful or diskless on the given host.

        :param str node_name: Name of the node
        :return: True if assigned (diskful or diskless) on given host.
        :rtype: bool
        """
        return self.is_diskful(node_name) or self.is_diskless(node_name)

    def diskless_nodes(self):
        """
        Returns the host names of all diskless nodes.

        :return: Host names of diskless nodes.
        :rtype: list[str]
        """
        return [n for n in self._assignments.keys() if self.is_diskless(n)]

    def diskful_nodes(self):
        """
        Returns the host names of all diskful nodes.

        :return: Host names of diskful nodes.
        :rtype: list[str]
        """
        return [n for n in self._assignments.keys() if self.is_diskful(n)]

    def is_thin(self):
        """
        Returns if the used storage pool of the resource is thin.

        :return: True if storage pool used is thin.
        :rtype: bool
        """
        with self._get_connection() as lin:
            stor_pool_list = lin.storage_pool_list_raise(None, filter_by_stor_pools=[self.volumes[0].storage_pool_name])
            return stor_pool_list.storage_pools[0].is_thin()

    # no decorator! (could recreate)
    def _delete(self, node_name=None):
        reinit = False
        if node_name is None:
            node_name = 'ALL'  # for error msg
            rs = self._lin.resource_dfn_delete(self._linstor_name)
            reinit = True
            self.defined = False
        else:
            if not self.is_assigned(node_name):
                return True
            rs = self._lin.resource_delete(node_name, self._linstor_name)
            if socket.gethostname() == node_name:  # deleted on myself
                reinit = True

        if not Linstor.all_api_responses_no_error(rs):
            raise linstor.LinstorError('Could not delete resource {} on node {}: {}'
                                       .format(self, node_name, Linstor.filter_api_call_response_errors(rs)[0]))
        if reinit:
            self._volumes = _VolumeDict()

        return self.__update_volumes()

    # no decorator! (could recreate)
    def delete(self, node_name=None, snapshots=True):
        """
        Deletes the resource globally or on the given host.

        If the node name is None, deletes the resource globally.

        :param str node_name: Deletes resource only from the specified node.
        :param bool snapshots: If True deletes snapshots prior deleting the resource

        :return: True if success, else raises LinstorError
        """
        if self._linstor_name is None:
            return True  # resource doesn't exist

        with self._get_connection() as lin:
            self._lin = lin

            if snapshots and node_name is None:  # only remove snapshots if resource definition will be deleted
                snapshot_list = lin.snapshot_dfn_list()[0]  # type: linstor.responses.SnapshotResponse
                for snap in [x for x in snapshot_list.snapshots if x.rsc_name.lower() == self._linstor_name.lower()]:
                    lin.snapshot_delete(rsc_name=self._linstor_name, snapshot_name=snap.snapshot_name)

            return self._delete(node_name)

    def drbd_proxy_enable(self, node_name_a, node_name_b):
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        proxy_enable_replies = self._lin.drbd_proxy_enable(self._linstor_name, node_name_a, node_name_b)
        if not Linstor.all_api_responses_no_error(proxy_enable_replies):
            raise linstor.LinstorError(
                'Could not enable drbd-proxy for resource {} between {} and {}: {}'.format(
                    self,
                    node_name_a,
                    node_name_b,
                    Linstor.filter_api_call_response_errors(proxy_enable_replies)[0])
            )
        return True

    def drbd_proxy_disable(self, node_name_a, node_name_b):
        if self._linstor_name is None:
            raise linstor.LinstorError("Resource '{n}' doesn't exist.".format(n=self.name))

        proxy_disable_replies = self._lin.drbd_proxy_disable(self._linstor_name, node_name_a, node_name_b)
        if not Linstor.all_api_responses_no_error(proxy_disable_replies):
            raise linstor.LinstorError(
                'Could not disable drbd-proxy for resource {} between {} and {}: {}'.format(
                    self,
                    node_name_a,
                    node_name_b,
                    Linstor.filter_api_call_response_errors(proxy_disable_replies)
                )
            )
        return True
